"""
>>> model = LogisticRegression()
>>> X = df['name'].str.len()
>>> df['len'] = df['name'].str.len()
>>> X = df[['len']]
>>> y = df['sex']
"""

"""
>>> model.fit(X[istrain], y[istrain])
LogisticRegression()
>>> model.score(X[istrain], y[istrain])
0.560972925379705
>>> model.score(X[~istrain], y[~istrain])
0.5492341356673961
"""

"""
>>> model.fit(X[istrain], y[istrain], sample_weight=df['count'][istrain])
LogisticRegression()

>>> model.score(X[istrain], y[istrain])
0.5173893902707462
>>> model.score(X[~istrain], y[~istrain])
0.5361050328227571

>>> model.score(X[istrain], y[istrain], sample_weight=df['count'][istrain])
0.5511512294352704
>>> model.score(X[~istrain], y[~istrain], sample_weight=df['count'][~istrain])
0.5972233131725709
"""

# neither year nor len are statistically significant predictors of sex
from pathlib import Path
import numpy as np
import pandas as pd
DATA_DIR = Path('.nlpia2-data')
df = pd.read_csv(DATA_DIR / 'baby-names-region.csv.gz')

np.random.seed(451)
df = df.sample(10_000)
names = df['name'].unique()
istrain = np.random.rand(len(names)) < .9
istest = ~istrain
df.groupby('name')['istrain'] = istrain
groups = df.groupby('name')
dfistrain = pd.DataFrame([istrain]).T
dfistrain
dfistrain.columns = ['istrain']
sum(dfistrain)
dfistrain.sum()
dfistrain.index = names
df.merge(dfistrain, on='name', kind='left')
df.merge?
df.merge?
df.merge(dfistrain, on='name')
dfistrain?
dfistrain
dfistrain['name'] = names
df.merge(dfistrain, on='name')
hist

istrain = df.merge(dfistrain, on='name')['istrain']
istrain
from sklearn.linear_model import LogisticRegression
model = LogisticRegression()
model.fit(df['name'].str.len()[istrain], df['sex'])
istrain = istrain.values
model.fit(df['name'].str.len()[istrain], df['sex'])
model.fit(df[['name']].str.len()[istrain], df['sex'])
X = df['name'].str.len()
df['len'] = df['name'].str.len()
X = df[['len']]
y = df['sex']
model.fit(X[istrain], y[istrain])
mmodel.score(X[istrain], y[istrain])
model.score(X[istrain], y[istrain])
model.score(X[~istrain], y[~istrain])
history -o -p
model.score(X[istrain], y[istrain], sample_weight=df['count'])
model.score(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model.fit(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model.score(X[istrain], y[istrain])
model.score(X[~istrain], y~[istrain])
model.score(X[~istrain], y[~istrain])
model.score(X[istrain], y[istrain], sample_weight=df['count'][istrain])
model.score(X[~istrain], y[~istrain], sample_weight=df['count'][~istrain])





from pathlib import Path
import pandas as pd
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression  # , Lasso
from tqdm import tqdm

DATA_DIR = Path('.nlpia2-data')
df = pd.read_csv(DATA_DIR / 'baby-names-region.csv.gz')

df = df.sample(1_000_000, random_state=1989)
np.random.seed(451)
istrain = np.random.rand(len(df)) < .9

vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
vectorizer.fit(df['name'][istrain])
vecs = vectorizer.transform(df['name'])

""" Use the count rather than freq in order to work with modern names

>>> model = LogisticRegression(class_weight='balanced', max_iter=10000)
>>> model.fit(vecs[istrain], df['sex'][istrain], sample_weight=df['count'][istrain])
LogisticRegression(class_weight='balanced', max_iter=10000)
>>> model.score(vecs[istrain], df['sex'][istrain], df['count'][istrain])
0.9826107512114707
>>> model.score(vecs[~istrain], df['sex'][~istrain], df['count'][~istrain])
0.983129802663326
>>> model.score(vecs[~istrain], df['sex'][~istrain])
0.9395149985982618
"""


model = LogisticRegression(class_weight='balanced', max_iter=10000)
model.fit(vecs[istrain], df['sex'][istrain], sample_weight=df['count'][istrain])
model.score(vecs[istrain], df['sex'][istrain], df['count'][istrain])
model.score(vecs[~istrain], df['sex'][~istrain], df['count'][~istrain])


from torch import nn, optim, manual_seed, tensor
import torch
from torch.utils.data import Dataset, DataLoader

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

# https://medium.com/deep-learning-study-notes/multi-layer-perceptron-mlp-in-pytorch-21ea46d50e62

# create custom dataset class


class CustomDataset(Dataset):

    def __init__(self, X, y):
        self.vectors = X
        self.labels = y

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, idx):
        label = tensor(self.labels[idx],
                       dtype=torch.float32)  # or torch.long or torch.float64
        vector = tensor(np.asarray(self.vectors[idx].todense()).squeeze(),
                        dtype=torch.long)
        return vector, label


NUM_FEATURES = len(vectorizer.get_feature_names())
BATCH_SIZE = 256 * 256

# https://medium.com/biaslyai/pytorch-linear-and-logistic-regression-models-5c5f0da2cb9


class NeuralLogisticRegression(torch.nn.Module):
    loss = torch.nn.BCELoss(size_average=True)

    def __init__(self):
        super().__init__()
        self.linear = torch.nn.Linear(NUM_FEATURES, 1)

    def forward(self, x):
        y_pred = F.sigmoid(self.linear(x))
        return y_pred


model = NeuralLogisticRegression()
loss_fun = NeuralLogisticRegression.loss


# for epoch in tqdm(range(20)):
#     model.train()
#     optimizer.zero_grad()
#     # Forward pass
#     y_pred = model(x_data)
#     # Compute Loss
#     loss = loss_fun(y_pred, y_data)
#     # Backward pass
#     loss.backward()
#     optimizer.step()

manual_seed(1989)
dataset = CustomDataset(vecs, (df['sex'] == 'F').astype(int).values)
print(next(iter(dataset)))
trainloader = DataLoader(
    dataset, batch_size=BATCH_SIZE, shuffle=True, num_workers=1)

p.train()
epoch_losses = []
for epoch in tqdm(range(10)):
    losses = []
    for batch_num, input_data in tqdm(enumerate(trainloader)):
        optimizer.zero_grad()
        x, y = input_data
        x = x.to(device).float()
        y = y.to(device).long()

        output = p(x)
        loss = loss_function(output, y)
        loss.backward()
        losses.append(loss.item())

        optimizer.step()

        if not batch_num % 2000:
            print('\tEpoch %d | Batch %d | Loss %6.2f' % (epoch, batch_num, loss.item()))
    epoch_losses.append(sum(losses) / len(losses))
    print('Epoch {} | Loss {:6.2f}'.format(epoch + 1, epoch_losses[-1]))


class Perceptron(nn.Module):
    def __init__(self):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Linear(num_features, 10),
            nn.Linear(10, 2),
            # nn.SoftMax(2)
        )
        self.linear = nn.Linear(num_features, 2)

    def forward(self, x):
        out = self.linear(x)
        return out


manual_seed(1989)
dataset = CustomDataset(vecs, (df['sex'] == 'F').astype(int).values)
print(next(iter(dataset)))
trainloader = DataLoader(
    dataset, batch_size=32, shuffle=True, num_workers=4)

p = Perceptron()
loss_function = nn.CrossEntropyLoss()
optimizer = optim.Adam(p.parameters(), lr=4e-5)


p.train()
epoch_losses = []
for epoch in range(10):
    losses = []
    for batch_num, input_data in enumerate(trainloader):
        optimizer.zero_grad()
        x, y = input_data
        x = x.to(device).float()
        y = y.to(device).long()

        output = p(x)
        loss = loss_function(output, y)
        loss.backward()
        losses.append(loss.item())

        optimizer.step()

        if not batch_num % 200:
            print('\tEpoch %d | Batch %d | Loss %6.2f' % (epoch, batch_num, loss.item()))
    epoch_losses.append(sum(losses) / len(losses))
    print('Epoch {} | Loss {:6.2f}'.format(epoch + 1, epoch_losses[-1]))



    g = Digraph(name)
    g.attr(rankdir='LR')
from graphviz import Digraph
dir(g)
    g = Digraph(name)
    g.attr(rankdir='LR')
    g = Digraph('text-NLU-vector')
    g.attr(rankdir='LR')
g.edge('Text\n(natural language)', 'NLU\n(rules, patterns, encoder)')
g.edge('NLU\n(rules, patterns, encoder)', 'Vector\n(tensor, embedding, encoding)')
g.save('text-NLU-vector.svg')
ls -al
firefox text-NLU-vector.svg
!firefox text-NLU-vector.svg
g.save()
!firefox text-NLU-vector.svg
g.format('svg')
g.format?
g.format
g.format = 'svg'
g.save()
!firefox text-NLU-vector.svg
ls -al
g.save(format='svg')
g.save?
g.format?
g.attr?
g.render(filename='text-NLU-vector.svg', cleanup=1, view=0)
g.render(filename='text-NLU-vector', cleanup=1, view=0)
!firefox text-NLU-vector.svg
g.render(filename='text-NLU-vector', cleanup=1, view=0, format='png')
history -f 'text-NLU-vector.py'



""" Notes from mob programming session Wed Oct 27, 2021 5:30 Pacific
(email engineering@tangibleai.com for details)

You need to have nlpia2/src/nlpia2 in your python path (or your CWD)
  for this to work

See [README.md](https://gitlab.com/prosocialai/nlpia2/README.md) for details.

You can `pip install nlpia2`
OR
`git clone git@gitlab.com:prosocialai/nlpia2`
then `cd nlpia2/src/nlpia2`
"""

"""
>>> from tqdm import tqdm
>>> import pandas as pd
>>> import numpy as np
>>> from sklearn.feature_extraction.text import TfidfVectorizer
>>> from sklearn.linear_model import LogisticRegression
"""
from tqdm import tqdm
import pandas as pd
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression  # noqa


"""
>>> from nlpia2.constants import DATA_DIR
>>> DATA_DIR
PosixPath('.../.nlpia-data/')
"""
from nlpia2.constants import DATA_DIR


"""
>>> np.random.seed(451)
>>> df = pd.read_csv(DATA_DIR / 'baby-names-region.csv.gz')
>>> df.sample(3).T
         6139665   2565339   22297
region        WV        MD        AK
sex            F         F         M
year        1987      1954      1988
name    Brittani       Ida   Maxwell
count         10        18         5
freq    0.000003  0.000005  0.000001
"""
df = pd.read_csv(DATA_DIR / 'baby-names-region.csv.gz')
df.sample(3).T
#          6139665   2565339   22297
# region        WV        MD        AK
# sex            F         F         M
# year        1987      1954      1988
# name    Brittani       Ida   Maxwell
# count         10        18         5
# freq    0.000003  0.000005  0.000001


"""
>>> istrain = np.random.rand(len(df))
>>> istrain.sum() / len(istrain)
0.8997952854283825
"""
np.random.seed(451)
istrain = np.random.rand(len(df)) < .9
istrain.sum() / len(istrain)
# 0.8997952854283825

"""
>>> vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
>>> vectorizer.fit(tqdm(df['name'][istrain]))
>>> feature_names = list(vectorizer.get_feature_names_out())
>>> len(feature_names)
5964
>>> feature_names[:5]
['a', 'aa', 'aab', 'aac', 'aad']
"""
vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3))
vectorizer.fit(tqdm(df['name'][istrain]))
feature_names = list(vectorizer.get_feature_names_out())
len(feature_names)
# 5964
feature_names[:5]
# ['a', 'aa', 'aab', 'aac', 'aad']


"""
>>> vecs = vectorizer.transform(tqdm(df['name']))
>>> vecs
<6241373x5964 sparse matrix of type '<class 'numpy.float64'>'
    with 86718620 stored elements in Compressed Sparse Row format>
"""
vecs = vectorizer.transform(tqdm(df['name']))
vecs
# <6241373x5964 sparse matrix of type '<class 'numpy.float64'>'
#     with 86718620 stored elements in Compressed Sparse Row format>


"""
>>> model = LogisticRegression(C=1, max_iter=2000)
>>> model.fit(vecs[istrain], df['sex'][istrain], sample_weight=df['freq'][istrain])
LogisticRegression(C=1, max_iter=2000)
"""
model = LogisticRegression(C=1, max_iter=2000)
model.fit(vecs[istrain], df['sex'][istrain], sample_weight=df['freq'][istrain])
# LogisticRegression(max_iter=2000)


"""
>>> names = [
...     'Maria', 'Aditi', 'Jessica', 'Olessya', 'Una',
...     'Hanna', 'Winnie', 'Olessya', 'Sylvia',
...     'Vish', 'Mohammed', 'Jon', 'John', 'Ted',
...     'Meijke', 'Uzo', 'Kazuma', 'Kemal']
>>> pd.Series(model.predict(vectorizer.transform(names)), index=names)
Maria       F
Aditi       F
Jessica     F
Olessya     F
Una         F
Hanna       F
Winnie      F
Olessya     F
Sylvia      F
Vish        M
Mohammed    M
Jon         M
John        M
Ted         M
Meijke      M
Uzo         M
Kazuma      F
Kemal       F
"""
names = [
    'Maria', 'Aditi', 'Jessica', 'Olessya', 'Una',
    'Hanna', 'Winnie', 'Olessya', 'Sylvia',
    'Vish', 'Mohammed', 'Jon', 'John', 'Ted',
    'Meijke', 'Uzo', 'Kazuma', 'Kemal']
print(pd.Series(model.predict(vectorizer.transform(names)),
                index=names))
# Maria       F
# Aditi       F
# Jessica     F
# Olessya     F
# Una         F
# Hanna       F
# Winnie      F
# Olessya     F
# Sylvia      F
# Vish        M
# Mohammed    M
# Jon         M
# John        M
# Ted         M
# Meijke      M
# Uzo         M
# Kazuma      F
# Kemal       F
# dtype: object


"""
>>> vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3),
...                              lowercase=False)
>>> vectorizer.fit(tqdm(df['name'][istrain]))
TfidfVectorizer(analyzer='char', lowercase=False, ngram_range=(1, 3))
"""
vectorizer = TfidfVectorizer(analyzer='char', ngram_range=(1, 3),
                             lowercase=False)
print(vectorizer.fit(tqdm(df['name'][istrain])))
# TfidfVectorizer(analyzer='char', lowercase=False, ngram_range=(1, 3))


"""
>>> feature_names = list(vectorizer.get_feature_names_out())
>>> len(feature_names)
8624
>>> feature_names[:5]
['A', 'Aa', 'Aab', 'Aad', 'Aah']
"""
feature_names = list(vectorizer.get_feature_names_out())
print(len(feature_names))
# 8624
print(feature_names[:5])
# ['A', 'Aa', 'Aab', 'Aad', 'Aah']


"""
>>> vecs = vectorizer.transform(tqdm(df['name']))
>>> vecs
<6241373x8624 sparse matrix of type '<class 'numpy.float64'>'
    with 87648303 stored elements in Compressed Sparse Row format>
"""
vecs = vectorizer.transform(tqdm(df['name']))
vecs
# <6241373x8624 sparse matrix of type '<class 'numpy.float64'>'
#     with 87648303 stored elements in Compressed Sparse Row format>


"""
>>> model = LogisticRegression(C=1, max_iter=2000)
>>> model.fit(vecs[istrain], df['sex'][istrain],
...           sample_weight=df['freq'][istrain])
LogisticRegression(C=1, max_iter=2000)
"""
model = LogisticRegression(C=1, max_iter=2000)
model.fit(vecs[istrain], df['sex'][istrain],
          sample_weight=df['freq'][istrain])
# LogisticRegression(C=1, max_iter=2000)


"""
>>> names = [
...     'Maria', 'Aditi', 'Jessica', 'Una',
...     'Hanna', 'Winnie', 'Olessya', 'Sylvia',
...     'Vish', 'Mohammed', 'Jon', 'John', 'Ted',
...     'Meijke', 'Uzo', 'Kazuma', 'Kemal']
>>> pd.Series(model.predict(vectorizer.transform(names)), index=names)
Maria       F
Aditi       F
Jessica     F
Una         F
Hanna       F
Winnie      F
Olessya     M
Sylvia      F
Vish        M
Mohammed    M
Jon         M
John        M
Ted         M
Meijke      M
Uzo         M
Kazuma      F
Kemal       M
"""
names = [
    'Maria', 'Aditi', 'Jessica', 'Una',
    'Hanna', 'Winnie', 'Olessya', 'Sylvia',
    'Vish', 'Mohammed', 'Jon', 'John', 'Ted',
    'Meijke', 'Uzo', 'Kazuma', 'Kemal']
print(pd.Series(model.predict(vectorizer.transform(names)),
                index=names))


"""
>>> (df['name'] == 'Kazuma').sum()
0
>>> (df['name'] == 'Kaz').sum()
1
>>> df[df['name'] == 'Kaz']
        region sex  year name  count      freq
2424961     LA   M  2015  Kaz      5  0.000002
"""
(df['name'] == 'Kazuma').sum()
0
(df['name'] == 'Kaz').sum()
1
df[df['name'] == 'Kaz']
#         region sex  year name  count      freq
# 2424961     LA   M  2015  Kaz      5  0.000002


"""
>>> (df['name'].str.startswith('Kaz')).sum()
151
>>> iskaz = df['name'].str.startswith('Kaz')
>>> df[iskaz].sort_values('freq')
        region sex  year      name  count      freq
4225691     NY   M  2006      Kazi      5  0.000001
4225692     NY   M  2006    Kazuki      5  0.000001
562866      CA   F  2005  Kazandra      5  0.000001
4223685     NY   M  2005      Kazi      5  0.000001
5360426     TX   F  2000  Kazandra      5  0.000001
...        ...  ..   ...       ...    ...       ...
625232      CA   M  1918     Kazuo     22  0.000011
626674      CA   M  1921     Kazuo     25  0.000012
623394      CA   M  1913     Kazuo     13  0.000013
627194      CA   M  1922     Kazuo     28  0.000013
399285      CA   F  1927    Kazuko     31  0.000014
[151 rows x 6 columns]
"""


"""
>>> model.coef_
array([[-5.56584447e-01, 5.14287250e-02, 2.69276390e-06, ...,
        -1.39283459e-04, -1.21600561e-04, -9.11513939e-05]])
>>> coef = pd.Series(model.coef_[0], index=feature_names)
>>> coef
A     -0.143835
Aa     0.033721
Aab    0.000001
Aad    0.000214
Aah    0.000008
         ...
zze   -0.000098
zzi   -0.001747
zzl   -0.000139
zzm   -0.000120
zzy   -0.000094
Length: 8624, dtype: float64
>>> coef.sort_values().round(2)
a     -0.50
i     -0.34
Mar   -0.30
y     -0.27
Ma    -0.25
       ...
mes    0.27
cha    0.27
J      0.28
am     0.32
o      0.40
"""
coef = pd.Series(model.coef_[0], index=feature_names)
print(coef.sort_values().round(2))
# a     -0.50
# i     -0.34
# Mar   -0.30
# y     -0.27
# Ma    -0.25
#        ...
# mes    0.27
# cha    0.27
# J      0.28
# am     0.32
# o      0.40


"""
>>> kazvec = vectorizer.transform(['Kazuma'])
>>> kazvec
<1x8624 sparse matrix of type '<class 'numpy.float64'>'
    with 14 stored elements in Compressed Sparse Row format>
>>> kazvec = pd.Series(kazvec.toarray()[0], index=feature_names)
>>> kazvec = kazvec[kazvec.abs() > 0]
>>> kazvec
K      0.126702
Ka     0.154755
Kaz    0.390671
a      0.100628
az     0.235593
azu    0.398854
m      0.115413
ma     0.159887
u      0.114679
um     0.248047
uma    0.312667
z      0.175519
zu     0.341750
zum    0.473007
"""
# TFIDFVectorizer.transform creates a sparse matrix, 1 row per input str
kazvec = vectorizer.transform(['Kazuma'])
print(kazvec)
# <1x8624 sparse matrix of type '<class 'numpy.float64'>'
#     with 14 stored elements in Compressed Sparse Row format>
pd.Series(kazvec.toarray()[0], index=feature_names)
kazvec = pd.Series(kazvec.toarray()[0], index=feature_names)
kazvec = kazvec[kazvec.abs() > 0]
kazvec
# K      0.126702
# Ka     0.154755
# Kaz    0.390671
# a      0.100628
# az     0.235593
# azu    0.398854
# m      0.115413
# ma     0.159887
# u      0.114679
# um     0.248047
# uma    0.312667
# z      0.175519
# zu     0.341750
# zum    0.473007

"""
>>> np.array(kazvec[kazvec != 0])[0]
array([0.10064136, 0.24160243, 0.36640437, 0.11589853, 0.1563424,
       0.40920386, 0.09844366, 0.12768782, 0.12058102, 0.26053805,
       0.32475817, 0.17116995, 0.31639993, 0.49863121])
>>> coef
a - 0.556584
aa     0.051429
aab    0.000003
aac    0.021992
aad    0.000415
    ...
zze - 0.000103
zzi - 0.001919
zzl - 0.000139
zzm - 0.000122
zzy - 0.000091
Length: 5986, dtype: float64
>>> coef.values[np.array(kazvec[kazvec != 0])[0] > 0]
>>> coef.values[kazvec != 0]
>>> type(kazvec != 0)
scipy.sparse.csr.csr_matrix
>>> (kazvec != 0).todense()
matrix([[True, False, False, ..., False, False, False]])
>>> np.array((kazvec != 0).todense())[0]
array([True, False, False, ..., False, False, False])
>>> coef[np.array((kazvec != 0).todense())[0] != 0]
a - 0.556584
az - 0.022166
azu - 0.000366
k      0.066347
ka - 0.182166
kaz    0.000030
m - 0.031816
ma - 0.280481
u      0.029682
um - 0.008430
uma    0.000764
z - 0.066873
zu - 0.001613
zum    0.000006
dtype: float64
>>> model.predict(kazvec)
array(['F'], dtype=object)
>>> model.predict_proba(kazvec)
array([[0.51219282, 0.48780718]])
>>> pd.Series(model.predict_proba(vectorizer.transform(names))[:, 1], index=names)
Maria       0.381240
Aditi       0.477269
Jessica     0.483249
Olessya     0.499430
Una         0.481148
Hanna       0.426783
Winnie      0.484074
Olessya     0.499430
Sylvia      0.484732
Vish        0.516631
Mohammed    0.569390
Jon         0.619116
John        0.678048
Ted         0.546004
Kazuma      0.487807
Meijke      0.528277
Kemal       0.494326
dtype: float64
>>> model.classes_
array(['F', 'M'], dtype=object)
>>> hist
>>> hist - o - p - f kemal_kazuma_labeled_female.md
"""

"""
>>> import yaml
>>> women = yaml.full_load(
...     str(DATA_DIR / 'women-in-ai-ethics-and-prosocial-ai.yml'))
>>> list(women)[:3]
['Vinita Silaparasetty', 'Ivana Bartoletti', 'Karolyn Gainfort']
>>> women = [w.split() for w in women]
>>> women = [w[1] if w[0] == 'Dr.' else w[0] for w in women]
>>> women[:6]
['Vinita', 'Ivana', 'Karolyn', 'Bulbul', 'Moojan', 'Hanan']
>>> pd.Series(model.predict(vectorizer.transform(women)), index=women)
Vinita     F
Ivana      F
Karolyn    F
Bulbul     M
Moojan     M
Hanan      F
Nikita     F
Zeynep     F
Melanie    F
Kathryn    F
Helen      F
Susan      F
Joy        M
Maria      F
"""



import os
import sys
import subprocess
from pathlib import Path
import json
import logging

import xmltodict
from xml.parsers.expat import ExpatError


log = logging.getLogger(__name__)

CODE_DIR = Path(__file__).resolve().absolute()
for i in range(4):
    if CODE_DIR.name in ['code', 'nlpia2']:
        break
    # print(f"NOT code dir: {CODE_DIR}")
    CODE_DIR = CODE_DIR.parent

REPO_DIR = CODE_DIR.parent
for i in range(4):
    if REPO_DIR.name in ['nlpia-manuscript', 'nlpia2']:
        break
    # print(f"not repo dir: {REPO_DIR}")
    REPO_DIR = REPO_DIR.parent

HOME_CODE_DIR = REPO_DIR.parent.parent
print(HOME_CODE_DIR)
assert HOME_CODE_DIR.name == 'code'
MANUSCRIPT_DIR = HOME_CODE_DIR / 'tangibleai' / 'nlpia-manuscript' / 'manuscript'
assert MANUSCRIPT_DIR.is_dir()
IMAGE_DIR = MANUSCRIPT_DIR / 'images'
assert IMAGE_DIR.is_dir()
SCRIPT_WORKING_DIR = os.getcwd()


def render_adoc(doctype='book', backend='html5', destination_dir='html', embedded=False, adoc_path='adoc/*.adoc'):
    """ Render adoc files in manuscript/asc to HTML or other viewable/printable format

    Input:
      backend (str): html5 xhtml5 docbook5 manpage
      doctype (str): article book manpage inline
      embedded (bool): whether to suppress enclosing document structure
    """

    # exit_code = subprocess.call(cmd, shell=True)  # exit_code == 0 if successful

    command = f'asciidoctor -d {doctype} -b {backend} -D {destination_dir} {adoc_path}'.split()
    return run(command=command, chdir=MANUSCRIPT_DIR)


def html2xml(destination_dir=None, html_path=None):
    """ Render adoc files in manuscript/asc to HTML or other viewable/printable format

    Implements `pandoc -t json -o ../xml/"$p".xml --indent=2 "$p"` in python.
    """
    # exit_code = subprocess.call(cmd, shell=True)  # exit_code == 0 if successful
    if html_path:
        htmlpaths = list(Path(html_path).glob(html_path))
    else:
        htmlpaths = list((MANUSCRIPT_DIR / 'xhtml5').glob('*.html'))

    if not destination_dir:
        destination_dir = MANUSCRIPT_DIR / 'xml'
    destination_dir.mkdir(exist_ok=True)

    command_messages = []
    for p in htmlpaths:
        # pandoc -t xml -o ../xml/"$p".xml 
        command = f'pandoc -t xml -o {str(destination_dir)} {p}'.split()
        output = run(command=command)
        messages.append(' '.join(command), output)
    return command_messages


def run(command, chdir=None):
    if chdir:
        log.info(f'Temporarily changing working directory to {chdir}')
        initial_cwd = os.getcwd()
        os.chdir(chdir)

    log.warning(f'Running: {" ".join(command)}')
    output = subprocess.run(command, capture_output=True)

    if chdir:
        os.chdir(initial_cwd)

    return {
        'stderr': output.stderr.decode("utf-8").splitlines(),
        'stdout': output.stderr.decode("utf-8").splitlines()
    }


def svg2png(filepath, dpi=300, width="100%", height="100%", background_color="white"):
    # exit_code = subprocess.call(cmd, shell=True)  # exit_code == 0 if successful
    filepath_noext = '.'.join(filepath.split('.')[:-1])
    # deprecated: cmd = f'inkscape --without-gui {filepath_noext}.svg -o {filepath_noext}.png'.split()
    cmd = str.split(f' inkscape {filepath_noext}.svg'
                    f' --export-filename {filepath_noext}.png'
                    f' --export-background={background_color}'
                    f' --export-dpi={dpi}'
                    # f'--export-width={width} --export-height={height}'
                    )
    return run(command=cmd, chdir=None)


def pprint_output(output, command=None):
    if command:
        print(f'\n{command} OUTPUT:')
    try:
        if output.get('stderr') or output.get('stdout'):
            json.dumps(str([output]), indent=4)
    except AttributeError:
        for cmd, outp in output:
            pprint_output(output=outp, command=cmd)


if __name__ == '__main__':
    
    # svgfilepaths = list(IMAGE_DIR.glob('**/*.svg'))
    svgfilepaths = [
        IMAGE_DIR / 'ch02' / 'survival-of-adequate-sentence-diagram.svg',
        ]
    svgfilepaths = [str(p) for p in svgfilepaths]    
    if len(sys.argv) >= 2:
        svgfilepaths = sys.argv[1:]

    flags = {'x': True, 'j': True}
    for flag in '-x -j --xml --json'.split():
        if flag in svgfilepaths:
            f = flag.lstrip('-')[:1]
            flags[f] = True
            del svgfilepaths[svgfilepaths.index(flag)]
    log.warning(f'flags = {flags}')

    for filepath in svgfilepaths:
        output = svg2png(filepath=filepath)
        pprint_output(output, command='render_html()')

    output = render_adoc(
        doctype='book',
        backend='html5',
        embedded=False)
    pprint_output(output, command='render_adoc()')

    # XML required for json
    if flags['x'] or flags['j']:
        print('render_html(backend=docbook) STDOUT:')
        output = render_adoc(
            doctype='book', 
            backend='docbook', 
            destination_dir='xml', 
            embedded=False
            )
        pprint_output(output)

    if flags['j']:
        bookdict = {}
        for filepath in (MANUSCRIPT_DIR / 'xml').glob('*.xml'):
            log.debug(f'Parsing {filepath}...')

            xml_text = open(filepath).read()
            d = None
            try:
                d = xmltodict.parse(xml_text)
            except ExpatError as e:
                log.error(f'Unable to parse xml in {filepath}:\n{e}')
                continue
            basename = filepath.with_suffix('').name
            bookdict[basename] = d
            filepath = MANUSCRIPT_DIR / 'json' / (basename + '.json')
            fout = open(filepath, 'wt')
            json.dump(d, fout, indent=4)
        filepath = MANUSCRIPT_DIR / 'json' / 'book.json'
        json.dump(d, fout, indent=4)



import os
from graphviz import Digraph
import shutil
from pathlib import Path


CODE_DIR = Path(__file__).resolve().absolute()
for i in range(4):
    if CODE_DIR.name in ['code', 'nlpia2']:
        break
    # print(f"NOT code dir: {CODE_DIR}")
    CODE_DIR = CODE_DIR.parent

REPO_DIR = CODE_DIR.parent
for i in range(4):
    if REPO_DIR.name in ['nlpia-manuscript', 'nlpia2']:
        break
    # print(f"not repo dir: {REPO_DIR}")
    REPO_DIR = REPO_DIR.parent

HOME_CODE_DIR = REPO_DIR.parent.parent
print(HOME_CODE_DIR)
assert HOME_CODE_DIR.name == 'code'
MANUSCRIPT_DIR = HOME_CODE_DIR / 'tangibleai' / 'nlpia-manuscript' / 'manuscript'
assert MANUSCRIPT_DIR.is_dir()
IMAGE_DIR = MANUSCRIPT_DIR / 'images'
assert IMAGE_DIR.is_dir()
SCRIPT_WORKING_DIR = os.getcwd()

print('IMAGE_DIR:')
print(IMAGE_DIR)


def draw_nlu_diagram(name='text-NLU-vector-graphviz', center_name='NLU\n(rules, patterns or encoder)'):
    g = Digraph(name, engine='dot')  # dot neato fdp sfdp
    g.attr(rankdir='LR')
    g.attr('node', shape='box')
    g.node(center_name)
    g.attr('node', shape='plaintext')
    g.edge('Text\n(natural language)', center_name)
    g.edge(center_name, 'Vector\n(numbers, tensor, embedding, encoding)')
    # g.save()

    g.render(filename=name, cleanup=1, view=0, format='png')

    dest = IMAGE_DIR / Path('ch01') / (name + '.png')

    print('Destination NLU diagram path:')
    print(dest)
    try:
        dest.resolve().absolute().unlink()
    except FileNotFoundError:
        pass
    shutil.move(name + '.png', str(dest.resolve().absolute()))
    print(dest)
    return g

# g.render(filename=name, cleanup=1, view=0, format='pdf')
# g.render(filename=name, cleanup=1, view=0, format='svg')
# !firefox text-NLU-vector.svg


def draw_nlg_diagram(name='vector-NLG-text-graphviz', center_name='NLG\n(rules, templates or decoder)'):
    g = Digraph(name)
    g.attr(rankdir='LR')
    g.attr('node', shape='box')
    g.node(center_name)
    g.attr('node', shape='plaintext')
    g.edge('Vector\n(numbers, tensor, embedding, encoding)', center_name)
    g.edge(center_name, 'Text\n(natural language)')

    # Creates "vector-NLG-text-graphviz.gv" file in working directory (.dot format graph)
    # g.save()

    g.render(filename=name, cleanup=1, view=0, format='png')

    # g.render(filename=name, cleanup=1, view=0, format='pdf')
    # g.render(filename=name, cleanup=1, view=0, format='svg')

    dest = IMAGE_DIR / Path('ch01') / (name + '.png')
    try:
        dest.resolve().absolute().unlink()
    except FileNotFoundError:
        pass
    shutil.move(name + '.png', str(dest.resolve().absolute()))

    print('NLG diagram destination path:')
    print(dest)
    return g


if __name__ == '__main__':
    draw_nlu_diagram()
    draw_nlg_diagram()



# import numpy as np
import pandas as pd
# import re
from pathlib import Path

from sklearn.linear_model import LogisticRegression
# from pandas.api.types import is_numeric_dtype

import seaborn as sns
from matplotlib import pyplot as plt
# import pandas as pd

sns.set_theme(color_codes=True)
plt.rcParams['figure.figsize'] = [6, 4.5]
plt.rcParams['figure.dpi'] = 120

# THIS_DIR = Path(__file__).expanduser().resolve().absolute().parent
# DATA_DIR = Path(THIS_DIR).parent.parent / '.nlpia2-data'
DATA_DIR = Path('/home/hobs/code/prosocialai/nlpia2/.nlpia2-data')
MANUSCRIPT_DIR = Path('/home/hobs/code/tangibleai/nlpia-manuscript')
CH05_IMAGES_DIR = MANUSCRIPT_DIR / 'manuscript' / 'images' / 'ch05'
DATAFILE = DATA_DIR / 'fake_vs_real.csv.gz'  # 'all.csv.gz'
assert CH05_IMAGES_DIR.is_dir()

df = pd.read_csv(DATAFILE, index_col=None)
df['title_len'] = df['title'].str.len()
# feature_names = ['title_len', 'text_len', 'title_allcap_ratio', 'title_allcap_token_len']
feature_names = ['title_len']
target_name = 'isfake'

X = df[feature_names]
y = df[target_name]

model = LogisticRegression(class_weight='balanced')
model.fit(X[feature_names], y)
print(model.score(X[feature_names], y))


plt.figure()
ax = sns.regplot(data=df.sample(200), x=feature_names[0], y=target_name,
                 logistic=True)
plt.xlabel('Title Length (characters)')
plt.ylabel('Fake News')
print(Path(__file__).name)
image_filepath = CH05_IMAGES_DIR / (Path(__file__).name[:-2] + 'png')
print(f'Saving logistic regression plot to {image_filepath}')
plt.savefig(image_filepath)
plt.show()



import os
from graphviz import Graph
import shutil
from pathlib import Path

print('HELLO')
CODE_DIR = Path(__file__).resolve().absolute()
for i in range(4):
    if CODE_DIR.name in ['code', 'nlpia2']:
        break
    CODE_DIR = CODE_DIR.parent

REPO_DIR = CODE_DIR.parent
for i in range(4):
    if REPO_DIR in ['nlpia-manuscript', 'nlpia2']:
        break
    REPO_DIR = REPO_DIR.parent

HOME_CODE_DIR = REPO_DIR.parent.parent
print(f'HOME_CODE_DIR: {HOME_CODE_DIR}')
assert HOME_CODE_DIR.name == 'code'
MANUSCRIPT_DIR = HOME_CODE_DIR / 'tangibleai' / 'nlpia-manuscript' / 'manuscript'
assert MANUSCRIPT_DIR.is_dir()
IMAGE_DIR = MANUSCRIPT_DIR / 'images'
assert IMAGE_DIR.is_dir()
SCRIPT_WORKING_DIR = os.getcwd()

print('IMAGE_DIR:')
print(IMAGE_DIR)

# TODO: get this text from the .yml file in nlpia-manuscript/code/data/ or qary/src/qary/data/nlpia/
BOOK_THIEF_TEXT = ("Reading 'The Shoulder Shrug' between two and three o'clock each morning, "
                   "post-nightmare, or during the afternoon, in the basement.")
BOOK_THIEF_TEXT = ("Trust me, though, the words were on their way, and when "
                   "they arrived, Liesel would hold them in her hands like "
                   "the clouds, and she would wring them out, like the rain.")


def get_text_bigrams(text=BOOK_THIEF_TEXT, tokenizer=str.split, num_tokens=8):
    tokens = tokenizer(text)
    return list(zip(tokens[:-1], tokens[1:]))[:num_tokens]


def draw_text_tokens(edges, name='draw-text-tokenx-graphviz', formats=['png', 'svg']):
    print('edges: ')
    print(edges)
    g = Graph(name)
    g.attr(rankdir='LR')
    g.attr('node', shape='box')
    for e in edges:
        g.edge(e[0], e[1])
    for f in formats:
        destfilename = f'{name}.{f}'
        g.render(filename=name, cleanup=1, view=0, format=f)
        dest = IMAGE_DIR / Path('ch02') / destfilename
        print('Destination path for draw_text_tokens():')
        print(dest)
        try:
            dest.resolve().absolute().unlink()
            print('overwriting existing file')
        except FileNotFoundError:
            print('creating new file')
        shutil.move(destfilename, str(dest.resolve().absolute()))
    return g

# !firefox text-NLU-vector.svg


if __name__ == '__main__':
    bigrams = get_text_bigrams(BOOK_THIEF_TEXT)
    draw_text_tokens(edges=bigrams, name='book-thief-split')



""" Character-based RNN from scratch with pytorch

References:
    - https://blog.floydhub.com/a-beginners-guide-on-recurrent-neural-networks-with-pytorch/
"""
import torch
from torch import nn

import numpy as np


texts = TEXTS
# ['hey how are you\n', 'good i am fine\n', 'have a nice day\n', 'i am great\n']
chars = set(''.join(texts))
int2char = dict(enumerate(chars))
char2int = {char: ind for ind, char in int2char.items()}
maxlen = len(max(texts, key=len))


for i in range(len(texts)):
    while len(texts[i]) < maxlen:
        texts[i] += ' '


input_seq = []
target_seq = []

for i in range(len(texts)):
    input_seq.append(texts[i][:-1])
    target_seq.append(texts[i][1:])
    print("Input Sequence: {}\nTarget Sequence: {}".format(input_seq[i], target_seq[i]))


for i in range(len(texts)):
    input_seq[i] = [char2int[character] for character in input_seq[i]]
    target_seq[i] = [char2int[character] for character in target_seq[i]]


dict_size = len(char2int)
seq_len = maxlen - 1
batch_size = len(texts)


def one_hot_encode(sequence, dict_size, seq_len, batch_size):
    # Creating a multi-dimensional array of zeros with the desired output shape
    features = np.zeros((batch_size, seq_len, dict_size), dtype=np.float32)

    # Replacing the 0 at the relevant character index with a 1 to represent that character
    for i in range(batch_size):
        for u in range(seq_len):
            features[i, u, sequence[i][u]] = 1
    return features


# Input shape --> (Batch Size, Sequence Length, One-Hot Encoding Size)
input_seq = one_hot_encode(input_seq, dict_size, seq_len, batch_size)


input_seq = torch.from_numpy(input_seq)
target_seq = torch.Tensor(target_seq)


# torch.cuda.is_available() checks and returns a Boolean True if a GPU is available, else it'll return False
is_cuda = torch.cuda.is_available()

# If we have a GPU available, we'll set our device to GPU. We'll use this device variable later in our code.
if is_cuda:
    device = torch.device("cuda")
    print("GPU is available")
else:
    device = torch.device("cpu")
    print("GPU not available, CPU used")


class Model(nn.Module):
    def __init__(self, input_size, output_size, hidden_dim, n_layers):
        super().__init__()

        # Defining some parameters
        self.hidden_dim = hidden_dim
        self.n_layers = n_layers

        # RNN Layer
        self.rnn = nn.RNN(input_size, hidden_dim, n_layers, batch_first=True)
        # Fully connected layer
        self.fc = nn.Linear(hidden_dim, output_size)

    def forward(self, x):

        batch_size = x.size(0)

        # Initializing hidden state for first input using method defined below
        hidden = self.init_hidden(batch_size)

        # Passing in the input and hidden state into the model and obtaining outputs
        out, hidden = self.rnn(x, hidden)

        # Reshaping the outputs such that it can be fit into the fully connected layer
        out = out.contiguous().view(-1, self.hidden_dim)
        out = self.fc(out)

        return out, hidden

    def init_hidden(self, batch_size):
        # This method generates the first hidden state of zeros which we'll use in the forward pass
        # We'll send the tensor holding the hidden state to the device we specified earlier as well
        hidden = torch.zeros(self.n_layers, batch_size, self.hidden_dim)
        return hidden


# Instantiate the model with hyperparameters
model = Model(input_size=dict_size, output_size=dict_size, hidden_dim=12, n_layers=1)
# We'll also set the model to the device that we defined earlier (default is CPU)
model.to(device)

# Define hyperparameters
n_epochs = 100
lr = 0.01

# Define Loss, Optimizer
criterion = nn.CrossEntropyLoss()
optimizer = torch.optim.Adam(model.parameters(), lr=lr)


# Training Run
for epoch in range(1, n_epochs + 1):
    optimizer.zero_grad()  # Clears existing gradients from previous epoch
    input_seq.to(device)
    output, hidden = model(input_seq)
    loss = criterion(output, target_seq.view(-1).long())
    loss.backward()  # Does backpropagation and calculates gradients
    optimizer.step()  # Updates the weights accordingly

    if epoch % 10 == 0:
        print('Epoch: {}/{}.............'.format(epoch, n_epochs), end=' ')
        print("Loss: {:.4f}".format(loss.item()))


# This function takes in the model and character as arguments and returns the next character prediction and hidden state
def predict(model, character):
    # One-hot encoding our input to fit into the model
    character = np.array([[char2int[c] for c in character]])
    character = one_hot_encode(character, dict_size, character.shape[1], 1)
    character = torch.from_numpy(character)
    character.to(device)

    out, hidden = model(character)

    prob = nn.functional.softmax(out[-1], dim=0).data
    # Taking the class with the highest probability score from the output
    char_ind = torch.max(prob, dim=0)[1].item()

    return int2char[char_ind], hidden


EOS = '\n'

# This function takes the desired output length and input characters as arguments, returning the produced sentence


def sample(model, out_len=None, start='hey'):
    model.eval()  # eval mode
    start = start.lower()
    # First off, run through the starting characters
    chars = [ch for ch in start]
    size = 1024 - len(chars)
    if out_len is not None:
        size = out_len - len(chars)
    # Now pass in the previous characters and get a new one
    for ii in range(size):
        char, h = predict(model, chars)
        if char == EOS:
            break
        chars.append(char)

    return ''.join(chars)



from nessvec.indexers import Index
import pandas as pd
from nessvec.files import load_embeddings

df = load_embeddings('word2vec')
small = df.iloc[:500_000]
index = Index(small)
me = df.T['reckless_abandon']
aggro = df.T['aggressive'] + df.T['altruism']
print('reckless_abandon fearless altruism')
aggro_me = (me + aggro) / 3
print(index.get_nearest(aggro_me.values))
fearless = df.T['fearless'] / 2 + df.T['altruism'] / 2
print('fearless altruism')
print(index.get_nearest(fearless))
print("fearless altruism agressive good kindness")
print(
    index.get_nearest(index.get_doc_vector("fearless altruism agressive good kindness")))
print("fearless altruism agressive good")
print(index.get_nearest(index.get_doc_vector("fearless altruism agressive good")))
print("fearless generosity agressive good")
print(index.get_nearest(index.get_doc_vector("fearless generosity aggressive good", use_idf=False)))



from scipy.io import wavfile
samplerate, data = wavfile.read('1210secretmorzecode.wav')
ls -hal
ls 1*
who
samplerate
data
data = data.abs()
data = np.abs(data)
import numpy as np
data = np.abs(data)
import pandas as pd
df = pd.DataFrame()
df['x'] = data[:3*samplerate]
df['t'] = range(len(df)) / samplerate
df['t'] = np.arange(len(df)) / samplerate
df
df = df.iloc[::samplerate / 100]
df.iloc[np.arange(0, samplerate / 100, -1)]
np.arange?
df.iloc[np.arange(0, -1, samplerate / 100)]
df.iloc[np.arange(0, None, samplerate / 100)]
df
samplerate
df.iloc[np.arange(0, len(df), samplerate / 100)]
df = df.iloc[np.arange(0, len(df), samplerate / 100)]
df
df.set_index('t')
df = df.set_index('t')
df.plot()
from matplotlib import pyplot as plt
plt.show()
df = df / 255
df
dot = [0] * 3 + [1] * 3 + [0] * 3
dash = [0] * 3 + [1] * 3 + [0] * 3
dash = [0] * 3 + [1] * 1 + [0] * 3
df.argmin()
df['x'].argmin()
df['x'][7]
df['x'].iloc[7]
i0 = df['x'].argmin()
df['x'].iloc[i0:].argmax()
i1 = df['x'].iloc[i0:].argmax()
i2 = df['x'].iloc[i1:].argmin()
i2
i2 = df['x'].iloc[i1+1:].argmin()
i2
i1 = df['x'].iloc[i0:].argmax()
i1
i2 = df['x'].iloc[i0+i1:].argmin()
i2
i1 = i0 + df['x'].iloc[i0:].argmax()
i2 = i1 + df['x'].iloc[i1:].argmin()
i3 = i2 + df['x'].iloc[i2:].argmax()
i1 - i0
i2 - i1
i3 - i2
dot = [0] * 6 + [1] * 6 + [0] * 6
dash = [0] * 6 + [1] * 6 * 3 + [0] * 6
df.convolve(dot)
df.x.convolve(dot)
df.x.values.convolve(dot)
np.convolve(df.x, dot)
df['dot'] = np.convolve(df.x, dot)
df['dot'] = [0] * (len(dot) / 2) + list(np.convolve(df.x, dot)) + [0] * (len(dot) / 2)
df['dot'] = [0] * int(len(dot) / 2) + list(np.convolve(df.x, dot)) + [0] * int(len(dot) / 2)
padlen = int(len(dot) / 2)
padlen
len(dot)
df['dot'] = np.convolve(df.x, dot)[padlen:-padlen]
df['dot'] = np.convolve(df.x, dot)[padlen:-padlen+1]
df.plot()
plt.show()
df.plot()
plt.show()
df['x'] - 1
4 * (df['x'] - .75)
df['x'] = 4 * (df['x'] - .75)
df['dot'] = np.convolve(df.x, dot)[padlen:-padlen+1]
df.plot()
plt.show()
dot = [-1] * 6 + [1] * 6 + [-1] * 6
dash = [-1] * 6 + [1] * 6 * 3 + [-1] * 6
df['dot'] = np.convolve(df.x, dot)[padlen:-padlen+1]
df['dash'] = np.convolve(df.x, dash)[padlen:-padlen+1]
padlen
dotpadlen = int(len(dot) / 2)
dashpadlen = int(len(dash) / 2)
df['dash'] = np.convolve(df.x, dash)[dashpadlen:-dashpadlen+1]
df['dot'] = np.convolve(df.x, dot)[dotpadlen:-dotpadlen+1]
df.plot()
plt.show()
df.plot()
plt.grid('on')
plt.show()
np.convolve?
df['dot'] = np.convolve(df.x, dot, mode='same')
df['dash'] = np.convolve(df.x, dash, mode='same')
df.plot()
plt.grid('on')
plt.show()
dot = np.array([-1] * 6 + [1] * 6 + [-1] * 6)
dot = dot / dot.abs().sum()
dot = dot / np.abs(dot).sum()
dot
dash = np.array([-1] * 6 + [1] * 6 * 3 + [-1] * 6)
dot = dash / np.abs(dash).sum()
dot = np.array([-1] * 6 + [1] * 6 + [-1] * 6)
dot = dot / np.abs(dot).sum()
dash = dash np.abs(dash).sum()
dash = dash / np.abs(dash).sum()
dash
df['dash'] = np.convolve(df.x, dash, mode='same')
df['dot'] = np.convolve(df.x, dot, mode='same')
df.plot()
plt.grid('on')
plt.show()
df.plot(style=['_','.','-'])
plt.show()
df.plot(style=['-','.','_'])
plt.grid('on')
plt.show()
df.plot(style=['-','.','_'], linewidth=[.5, 2, 4])
df['x'].plot(style='-', linewidth=.75, alpha=.5)
df['dot'].plot(style='.', linewidth=2, alpha=.75)
df['dash'].plot(style='_', linewidth=5, alpha=.75)
plt.show()
df['x'].plot(linestyle='solid', linewidth=1, alpha=1)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.4)
df['dot'].plot(linestyle='dotted', linewidth=4, alpha=.35)
plt.show()
df['x'].plot(linestyle='solid', linewidth=2.5, alpha=1)
df['dash'].plot(linestyle='dashed', linewidth=2.5, alpha=.4)
df['dot'].plot(linestyle='dotted', linewidth=2.5, alpha=.35)
plt.grid('on')
plt.show()
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2.5, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=3, alpha=1)
plt.grid('on')
plt.show()
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.legend()
plt.show()
hist -f code/tangibleai/nlpia2/src/nlpia2/ch07/secret_message_convolved_line_plot.hist.py
hist -o -p -f code/tangibleai/nlpia2/src/nlpia2/ch07/secret_message_convolved_line_plot.hist.md
import seaborn
seaborn.set_theme()
hist
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.legend()
plt.show()
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.xlabel('Time (sec)')
plt.ylabel('Signal amplitude (-1 to +1)')
plt.legend()
plt.show()
seaborn.set_theme('whitegrid')
seaborn.set_theme('paper')
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.xlabel('Time (sec)')
plt.ylabel('Signal amplitude (-1 to +1)')
plt.legend()
plt.show()
seaborn.set_theme('talk')
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.xlabel('Time (sec)')
plt.ylabel('Signal amplitude (-1 to +1)')
plt.legend()
plt.show()
seaborn.set_theme('poster')
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.xlabel('Time (sec)')
plt.ylabel('Signal amplitude (-1 to +1)')
plt.legend()
plt.show()
seaborn.set_theme('whitegrid')
seaborn.set_theme('notebook')
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.xlabel('Time (sec)')
plt.ylabel('Signal amplitude (-1 to +1)')
plt.legend()
plt.show()
seaborn.set_style('whitegrid')
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.xlabel('Time (sec)')
plt.ylabel('Signal amplitude (-1 to +1)')
plt.legend()
plt.show()
fig = plt.figure(figsize=(700,300))
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.xlabel('Time (sec)')
plt.ylabel('Signal amplitude (-1 to +1)')
plt.legend()
plt.show()
fig = plt.figure(figsize=(10,4))
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.xlabel('Time (sec)')
plt.ylabel('Signal amplitude (-1 to +1)')
plt.legend()
plt.show()
fig = plt.figure(figsize=(10,7))
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.xlabel('Time (sec)')
plt.ylabel('Signal amplitude (-1 to +1)')
plt.legend()
plt.show()
fig = plt.figure(figsize=(10,6))
df['x'].plot(linestyle='solid', linewidth=2, alpha=.5)
df['dash'].plot(linestyle='dashed', linewidth=2, alpha=.75)
df['dot'].plot(linestyle='dotted', linewidth=2, alpha=1)
plt.grid('on')
plt.xlabel('Time (sec)')
plt.ylabel('Signal amplitude (-1 to +1)')
plt.legend()
plt.show()
hist
df.to_csv('secret_message_convolved_line_plot.csv')
ls
cd ~/code/tangibleai/nlpia-manuscript/manuscript/images/
df.to_csv('secret_message_convolved_line_plot.csv')
df2 = pd.read_csv('secret_message_convolved_line_plot.csv', index_col=0)
df
from matplotlib import pyplot as plt
from urllib.request import urlretrieve
f = urlretrieve(url, 'titanic-distress-signal-simulated.wav')
hist
PKG_DIR = Path('/home/hobs/code/tangibleai/nlpia2/src/nlpia2')
from pathlib import Path
PKG_DIR = Path('/home/hobs/code/tangibleai/nlpia2/src/nlpia2')
PKG_NAME = PKG_DIR.name
SRC_DIR = PKG_DIR.parent
REPO_DIR = SRC_DIR.parent
MANUSCRIPT_DIR = REPO_DIR.parent / 'nlpia-manuscript' / 'manuscript'
IMAGES_DIR = MANUSCRIPT_DIR / 'images'
ADOC_DIR = MANUSCRIPT_DIR / 'adoc'
MANUSCRIPT_DIR.is_dir()
IMAGES_DIR
IMAGES_DIR.is_dir()
pwd
ls -hal
mv secret_message_convolved_line_plot.csv /home/hobs/.nlpia2-data/
hist -o -p -f /home/hobs/code/tangibleai/nlpia2/src/nlpia2/ch07/secret_message_convolved_line_plot.hist.md
hist -f /home/hobs/code/tangibleai/nlpia2/src/nlpia2/ch07/secret_message_convolved_line_plot.hist.py



import logging
import math
import torch
import torch.nn as nn

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.WARNING)


class CNNTextClassifier(nn.ModuleList):

    def __init__(self, params=None, **kwargs):
        super().__init__()

        self.convolvers = []
        self.poolers = []

        self.seq_len = getattr(params, 'seq_len', 40)
        self.vocab_size = getattr(params, 'vocab_size', 3000)
        self.embedding_size = getattr(params, 'embedding_size', 64)
        self.kernel_lengths = list(getattr(params, 'kernel_lengths', [3, 4]))

        self.stride = getattr(params, 'stride', 2)
        self.strides = getattr(params, 'strides')
        if not self.strides:
            self.strides = [self.stride] * len(self.kernel_lengths)

        self.dropout_portion = getattr(params, 'dropout_portion', .2)
        self.dropout = nn.Dropout(self.dropout_portion)

        self.conv_output_size = getattr(params, 'conv_output_size', 32)
        self.__dict__.update(kwargs)

        for param_name, param_val in vars(self).items():
            if param_name.startswith('_'):
                continue
            if param_name in kwargs:
                setattr(self, param_name, kwargs[param_name])
            log.info(f'MODEL: {param_name}: {getattr(self, param_name)} ({type(getattr(self, param_name))})')

        self.embedding = nn.Embedding(self.vocab_size + 1, self.embedding_size, padding_idx=0)

        # default: 4 CNN layers with max pooling
        for i, (kernel_len, stride) in enumerate(zip(self.kernel_lengths, self.strides)):
            self.convolvers.append(nn.Conv1d(self.seq_len, self.conv_output_size, kernel_len, stride))
            # setattr(self, f'conv_{i + 1}', self.convolvers[i])
            self.poolers.append(nn.MaxPool1d(kernel_len, stride))
            # setattr(self, f'pool_{i + 1}', self.poolers[i])

        self.encoding_size = self.cnn_output_size()
        self.linear_layer = nn.Linear(self.encoding_size, 1)

    def cnn_output_size(self):
        """ Calculate the number of encoding dimensions output from CNN layers

        Convolved_Features = ((embedding_size + (2 * padding) - dilation * (kernel - 1) - 1) / stride) + 1
        Pooled_Features = ((embedding_size + (2 * padding) - dilation * (kernel - 1) - 1) / stride) + 1

        source: https://pytorch.org/docs/stable/generated/torch.nn.Conv1d.html
        """
        out_conv_1 = ((self.embedding_size - 1 * (self.kernel_lengths[0] - 1) - 1) / self.stride) + 1
        out_conv_1 = math.floor(out_conv_1)
        out_pool_1 = ((out_conv_1 - 1 * (self.kernel_lengths[0] - 1) - 1) / self.stride) + 1
        out_pool_1 = math.floor(out_pool_1)

        out_pool_total = 0
        for kernel_len, stride in zip(self.kernel_lengths, self.strides):
            out_conv = ((self.embedding_size - 1 * (kernel_len - 1) - 1) / stride) + 1
            out_conv = math.floor(out_conv)
            out_pool = ((out_conv - 1 * (kernel_len - 1) - 1) / stride) + 1
            out_pool = math.floor(out_pool)
            out_pool_total += out_pool

        # Returns "flattened" vector (input for fully connected layer)
        return out_pool_total * self.conv_output_size

    def forward(self, x):
        """ Takes sequence of integers (token indices) and outputs binary class label """

        x = self.embedding(x)

        conv_outputs = []
        for (conv, pool) in zip(self.convolvers, self.poolers):
            z = conv(x)
            z = torch.relu(z)
            z = pool(z)
            conv_outputs.append(z)

        # The output of each convolutional layer is concatenated into a unique vector
        union = torch.cat(conv_outputs, 2)
        union = union.reshape(union.size(0), -1)

        # The "flattened" vector is passed through a fully connected layer
        out = self.linear_layer(union)
        # Dropout is applied
        out = self.dropout(out)
        # Activation function is applied
        out = torch.sigmoid(out)

        return out.squeeze()



import yaml
yaml.full_load(open('data/quotes.yml'))
quotes = _
for q in quotes:
    print(q['text'])
sentences = []
for q in quotes:

    print(q['text'])
import spacy
nlp = spacy.load('en_core_web_md')
sentences = []
for q in quotes:

    txt = q['text']
    doc = nlp(txt)
    sentences.extend([s.text for s in doc.sents])
sentences
len(sentences)
import numpy as np
np.random.choose(sentences)
dir(np.random)
np.random.choice(sentences)
np.random.choice(sentences)
np.random.choice(sentences)
np.random.choice(sentences)
np.random.choice(sentences)
np.random.choice(sentences)
np.random.choice(sentences, prob=[1] + [0] * (len(sentences) - 1))
np.random.choice(sentences, probs=[1] + [0] * (len(sentences) - 1))
help(np.random.choice)
np.random.choice(sentences, p=[1] + [0] * (len(sentences) - 1))
sentences[0]
from collections import Counter
Counter(sentences[0])
sentences = []
for q in quotes:

    txt = q['text']
    doc = nlp(txt)
    sentences.extend([s.text for s in doc.sents])
sentences = [re.find_all(r'[a-z ]', s.lower()) for s in sentences]
import re
sentences = [re.find_all(r'[a-z ]', s.lower()) for s in sentences]
sentences = [re.findall(r'[a-z ]', s.lower()) for s in sentences]
sentences[0]
sentences = [['<SOS>'] + clist + ['<EOS>'] for clist in sentences]
sentences = [Counter(clist) for clist in sentences]
sentences[0]
sum(sentences)
type(sentences[0])
import pandas as pd
sentences[1]
sum(sentences[:2])
df = pd.DataFrame(sentences)
df.sum()
df.head()
prob = df.sum()
prob = prob / prob.sum()
prob
generated_sentence = []
while c != '<EOS>':
    c = np.random.choice(prob.index.values, p=prob.values)
    generated_sentence.append(c)
generated_sentence = []
c = '<SOS>'
while c != '<EOS>':
    c = np.random.choice(prob.index.values, p=prob.values)
    generated_sentence.append(c)
generated_sentence
''.join(generated_sentence)
generated_sentence = []
c = '<SOS>'
while c != '<EOS>':
    c = np.random.choice(prob.index.values, p=prob.values)
    generated_sentence.append(c)
''.join(generated_sentence)
generated_sentence = []
c = '<SOS>'
while c != '<EOS>':
    c = np.random.choice(prob.index.values, p=prob.values)
    generated_sentence.append(c)
''.join(generated_sentence)
generated_sentence = []
c = '<SOS>'
while c != '<EOS>':
    while c != '<SOS>':
        c = np.random.choice(prob.index.values, p=prob.values)
    generated_sentence.append(c)
generated_sentence = []
c = '<SOS>'
while c != '<EOS>':
    c = np.random.choice(prob.index.values, p=prob.values)
    if c == '<SOS>':
        continue
    generated_sentence.append(c)
generated_sentence
''.join(generated_sentence)


def generate_sentence(model_order=0):
    generated_sentence = []
    c = '<SOS>'
    while c != '<EOS>':
        c = np.random.choice(prob.index.values, p=prob.values)
        if c == '<SOS>':
            continue
        generated_sentence.append(c)
    return ''.join(generated_sentence)


generate_sentence()
generate_sentence()
generate_sentence()
generate_sentence()
hist


def generate_ngrams(seq):
    return list(zip(seq[:-1], seq[1:]))


generate_ngrams('Hello world')
hist
sentences
df_counts = pd.DataFrame(sentences)
df.head()

sentences = []
for q in quotes:

    txt = q['text']
    doc = nlp(txt)
    sentences.extend([s.text for s in doc.sents])
[generate_ngrams(s) for s in sentences]
sentences = [re.findall(r'[a-z ]', s.lower()) for s in sentences]
sentences = [['<SOS>'] + clist + ['<EOS>'] for clist in sentences]
sentence_2grams = list(generate_ngrams(s) for s in sentences)
sentence_2grams[0]
counts = {}
counts = dict(zip(probs.index.values, [] * len(probs)))
for s in sentence_2grams:
    for tg in s:
        counts[tg[0]] += [tg[1]]
counts = dict(zip(probs.index.values, [] * len(prob)))
for s in sentence_2grams:
    for tg in s:
        counts[tg[0]] += [tg[1]]
counts = dict(zip(prob.index.values, [] * len(prob)))
for s in sentence_2grams:
    for tg in s:
        counts[tg[0]] += [tg[1]]
prob.index.values
counts
counts = dict(zip(prob.index.values, [[]] * len(prob)))
for s in sentence_2grams:
    for tg in s:
        counts[tg[0]] += [tg[1]]
counts
counts = dict(zip(prob.index.values, [[]] * len(prob)))
for s in sentence_2grams:
    for tg in s:
        counts[tg[0]] += [tg[1]]
counts['s']
counts['s']
len(counts['s'])
len(counts['e'])
counts = dict(zip(prob.index.values, [[]] * len(prob)))
for s in sentence_2grams:
    for tg in s:
        print tg[0], tg[1]
        counts[tg[0]] += [tg[1]]
        break
counts = dict(zip(prob.index.values, [[]] * len(prob)))
for s in sentence_2grams:
    for tg in s:
        print(tg[0], tg[1])
        counts[tg[0]] += [tg[1]]
        break
counts = dict(zip(prob.index.values, [[]] * len(prob)))
for s in sentence_2grams:
    print(s)
    for tg in s:
        print(tg[0], tg[1])
        counts[tg[0]] += [tg[1]]
        break
counts = {}
for c in prob.index.values:
    counts[c] = list()
for s in sentence_2grams:
    print(s)
    for tg in s:
        print(tg[0], tg[1])
        counts[tg[0]].append(tg[1])
        break
counts
counts['SOS']
counts['<SOS>']
counters = dict()
for k, v in counts.items():
    counters[k] = Counter(v)
dfcounters = pd.DataFrame(counters)
dfcounters.head()
counts['<EOS>']
counts['t']
counts = {}
for c in prob.index.values:
    counts[c] = list()
for s in sentence_2grams:
    print(s)
    for tg in s:
        print(tg[0], tg[1])
        counts[tg[0]].append(tg[1])
counts['t']
counters = dict()
for k, v in counts.items():
    counters[k] = Counter(v)
dfcounters = pd.DataFrame(counters)
dfcounters.head()
dfcounters / dfcounters.sum()
dfcounters = dfcounters.fillna(0)
dfcounters['<EOS>']
dfcounters['<EOS>'].sum()
dfcounters['<EOS>'][' '] = 1
dfcounters.sum()
dfcounters / dfcounters.sum()
dfprobs = dfcounters / dfcounters.sum()
dfprobs.head()
import pandas as pd
from collections import Counter
import numpy as np


def generate_ngrams(seq):
    return list(zip([seq[:-1], seq[1:]]))


# def generate_ngrams(seq, n=2):
#     return list(zip([seq[i:-j] for i,j in zip(range(n-1), range(1,n))], seq[(n-1):]]))


def generate_sentence(model_order=0, probs=None):
    model_order = 0 if len(probs.shape) == 1 else 1
    # if model_order:
    #     model_order = len(probs.shape) - 1
    if model_order:
        generated_sentence = []
        c = '<SOS>'
        while c != '<EOS>':
            ctemp = np.random.choice(probs.index.values, p=probs[c].values)
            if ctemp == '<SOS>':
                continue
            c = ctemp
            generated_sentence.append(c)
        return ''.join(generated_sentence)
    else:
        generated_sentence = []
        c = '<SOS>'
        while c != '<EOS>':
            c = np.random.choice(probs.index.values, p=probs.values)
            if c == '<SOS>':
                continue
            generated_sentence.append(c)
        return ''.join(generated_sentence)


probs.head()
dfprobs.head()
generate_sentence(model_order=1, probs=dfprobs)
generate_sentence(model_order=1, probs=dfprobs)
generate_sentence(model_order=1, probs=dfprobs)
import pandas as pd
from collections import Counter
import numpy as np


def generate_ngrams(seq):
    return list(zip([seq[:-1], seq[1:]]))


# def generate_ngrams(seq, n=2):
#     return list(zip([seq[i:-j] for i,j in zip(range(n-1), range(1,n))], seq[(n-1):]]))


def generate_sentence(model_order=0, probs=None):
    model_order = 0 if len(probs.shape) == 1 else 1
    # if model_order:
    #     model_order = len(probs.shape) - 1
    if model_order:
        generated_sentence = []
        c = '<SOS>'
        while c != '<EOS>':
            ctemp = np.random.choice(probs.index.values, p=probs[c].values)
            if ctemp == '<SOS>':
                continue
            print(ctemp)
            c = ctemp
            generated_sentence.append(c)
        return ''.join(generated_sentence)
    else:
        generated_sentence = []
        c = '<SOS>'
        while c != '<EOS>':
            c = np.random.choice(probs.index.values, p=probs.values)
            if c == '<SOS>':
                continue
            generated_sentence.append(c)
        return ''.join(generated_sentence)


generate_sentence(model_order=1, probs=dfprobs)
hist - o - p - f quotes_language_model_firt_order_markhov_chain.ipy
hist - f quotes_language_model_firt_order_markhov_chain.py



import spacy
import os
from pathlib import Path
try:
    name = 'en_core_web_sm'
    nlp = spacy.load(name)
    print(f'SpaCy {name} is already installed.')
except (NameError, ImportError) as e:
    print(e)
    print(f'Downloading the **{name}** spacy language model...')
    spacy.cli.download(name)
    nlp = spacy.load(name)

modeldir = Path(nlp._path).parent.parent
files = os.listdir(str(modeldir))
# print('\n'.join(files));
name = 'en_core_web_md'
if any(f.startswith(name) for f in files):
    print(f'SpaCy {name} is installed here: {modeldir}/{name}*')
else:
    print(f"Couldn't find {name}\n  in {modeldir}\n  among {files}")
    print(f'Trying again to download **{name}** spacy language model...')
    spacy.cli.download(name)



import pandas as pd
import re
from pathlib import Path


DATA_DIR = Path('~/code/tangibleai/fake-news/data').expanduser().resolve().absolute()
df = pd.read_csv(Path(DATA_DIR) / 'all.csv.gz', index_col=None)

pattern = r'\w+(?:\'\w+)?|[^\w\s]'
df['title_len'] = df['title'].str.len()
df['title_num_tokens'] = df['title'].str.findall(pattern).apply(len)
df['title_ave_token_len'] = df['title_len'] / df['title_num_tokens']

wpat = r'\w+(?:\'\w+)?'
df['title_num_words'] = df['title'].str.findall(wpat).str.len()
df['title_wlen'] = df['title'].str.findall(wpat).str.join('').str.len()
df['title_ave_wordlen'] = df['title_wlen'] / df['title_num_words']

from sklearn.linear_model import LogisticRegression
model = LogisticRegression()
df = df.sample(len(df))  # <1>
# <1> Shuffle your data to improve convergence when your data is "stratified" by the target as it was here.

X = df[[c for c in df.columns if c.startswith('title_')]]
y = df['isfake']


def tokenize(s):
    pattern = r'\w+(?:\'\w+)?|[^\w\s]'
    return re.findall(pattern)


def tokenize_words(s):
    wordpat = r'\w+(?:\'\w+)?'
    return re.findall(wordpat)


def filter_nonwords(tokens):
    return [t for t in tokens if t and t[0]]


def _wordlen(tokens):
    return tokenize_words()



from collections import Counter
import numpy as np
import pandas as pd
from pathlib import Path
import re
import yaml

import spacy
nlp = spacy.load('en_core_web_md')


DATA_DIR = Path('../data').expanduser().resolve().absolute()


def generate_ngrams(seq):
    return list(zip([seq[:-1], seq[1:]]))


# def generate_ngrams(seq, n=2):
#     return list(zip([seq[i:-j] for i,j in zip(range(n-1), range(1,n))], seq[(n-1):]]))


def generate_sentence(probs=None):
    model_order = 0 if len(probs.shape) == 1 else 1
    # if model_order:
    #     model_order = len(probs.shape) - 1
    if model_order:
        generated_sentence = []
        c = '<SOS>'
        while c != '<EOS>':
            ctemp = np.random.choice(probs.index.values, p=probs[c].values)
            if ctemp == '<SOS>':
                continue
            c = ctemp
            generated_sentence.append(c)
        return ''.join(generated_sentence)
    else:
        generated_sentence = []
        c = '<SOS>'
        while c != '<EOS>':
            c = np.random.choice(probs.index.values, p=probs.values)
            if c == '<SOS>':
                continue
            generated_sentence.append(c)
        return ''.join(generated_sentence)


def tokenize(s):
    return re.findall(r'[a-z ]', s.lower())


def load_token_lists(filepath=DATA_DIR / 'quotes.yml', tokenizer=tokenize):
    quotes = yaml.full_load(open(filepath))
    sentences = []
    for q in quotes:
        txt = q['text']
        doc = nlp(txt)
        sentences.extend([s.text for s in doc.sents])
    # convert to lists of chars
    sentences = [tokenizer(s) for s in sentences]
    sentences = [['<SOS>'] + clist + ['<EOS>'] for clist in sentences]
    return sentences


def calc_probs(token_lists):
    counters = [Counter(clist) for clist in sentences]
    df = pd.DataFrame(counters)
    # >>> df.sum()
    # <SOS>     174.0
    # t        1302.0
    # h         651.0
    # e        1635.0
    #          2674.0
    # g         289.0
    # ...
    # q           6.0
    # z          13.0
    # dtype: float64

    # df.head()
    #    <SOS>     t    h     e          g     i  ...    y    w    b   x   j   q   z
    # 0      1  10.0  5.0  12.0  17.0  5.0  11.0  ...  NaN  NaN  NaN NaN NaN NaN NaN
    # 1      1   4.0  4.0  14.0  22.0  4.0  10.0  ...  4.0  1.0  2.0 NaN NaN NaN NaN
    # 2      1   7.0  NaN  15.0  11.0  2.0  10.0  ...  1.0  NaN  1.0 NaN NaN NaN NaN
    # 3      1  10.0  5.0  13.0  17.0  2.0   8.0  ...  NaN  1.0  1.0 NaN NaN NaN NaN
    # 4      1   7.0  3.0  15.0  22.0  4.0   7.0  ...  1.0  NaN  3.0 NaN NaN NaN NaN
    # [5 rows x 29 columns]
    prob = df.sum()
    prob = prob / prob.sum()
    # prob
    # <SOS>    0.010677
    # t        0.079892
    # h        0.039946
    # e        0.100325
    #          0.164079
    # g        0.017733
    # ...
    return prob


def calc_conditional_probs(ngram_lists):
    """ FIXME """
    beginning_counts, ending_counts = Counter(), Counter()
    for ngrams in ngram_lists:
        beginnings, endings = zip(*ngrams)
        beginning_counts += Counter(beginnings)
        ending_counts += Counter(endings)
    df_probs = pd.DataFrame(np.zeros(len(endings), len(beginnings)), columns=beginnings.keys(), index=endings.keys())
    return df_probs.fillna(0)


if __name__ == '__main__':
    sentences = load_token_lists()
    sentence_2grams = list(generate_ngrams(s) for s in sentences)
    probs = calc_probs([[ngram for ngram[-1] in s] for s in sentence_2grams])
    generate_sentence(probs=probs)
    probs = calc_conditional_probs(sentence_2grams)
    generate_sentence()
    generate_sentence()
    generate_sentence()



""" DEPRECATEd: use render_adoc.py and build.sh instead

FIXME: `cat` *.adoc | `asciidoctor` >> *.html | `pandoc` >> *.xml | `xmltodict` >> *.json

Convert render all adoc files in the manuscript/adoc/ dir to html then xml then json and load them into a list of dictionaries.

```bash
mkdir -p ../xml
mkdir -p ../html


for p in *.adoc ; do
  echo "path: $p";
  asciidoctor -d book -b html5 "$p" -o ../html/
done

for p in *.html ; do
  echo "path: $p";
  pandoc -t xml -o ../xml/"$p".xml --indent=2 "$p"
done
```
"""

from pathlib import Path
import logging

log = logging.getLogger(__name__)

CODE_DIR = Path(__file__).resolve().absolute()
for i in range(4):
    if CODE_DIR.name in ['code', 'nlpia2']:
        break
    # print(f"NOT code dir: {CODE_DIR}")
    CODE_DIR = CODE_DIR.parent

REPO_DIR = CODE_DIR.parent
for i in range(4):
    if REPO_DIR.name in ['nlpia-manuscript', 'nlpia2']:
        break
    # print(f"not repo dir: {REPO_DIR}")
    REPO_DIR = REPO_DIR.parent

HOME_CODE_DIR = REPO_DIR.parent.parent
# print(HOME_CODE_DIR)
# assert HOME_CODE_DIR.name == 'code'
MANUSCRIPT_DIR = HOME_CODE_DIR / 'tangibleai' / 'nlpia-manuscript' / 'manuscript'
# assert MANUSCRIPT_DIR.is_dir()
IMAGE_DIR = MANUSCRIPT_DIR / 'images'
# assert IMAGE_DIR.is_dir()
# SCRIPT_WORKING_DIR = os.getcwd()



import spacy
from spacy import displacy

REPO_DIR = Path(__file__).resolve().absolute().parent.parent.parent
IMAGE_DIR = REPO_DIR / 'manuscript' / 'images'

"""
>>> text = ("Trust me, though, the words were on their way, and when "
...         "they arrived, Liesel would hold them in her hands like "
...         "the clouds, and she would wring them out, like the rain.")
>>> tokens = text.split()
>>> tokens[:8]
['Trust', 'me,', 'though,', 'the', 'words', 'were', 'on', 'their']
>>> import re
>>> pattern = r'\w+(?:\'\w+)?|[^\w\s]'  # <1>
>>> texts = [text]
>>> texts.append("There's no such thing as survival of the fittest."
...              "Survival of the most adequate, maybe.")
>>> tokens = list(re.findall(pattern, texts[-1]))
>>> tokens[:8]
["There's", 'no', 'such', 'thing', 'as', 'survival', 'of', 'the']
>>> tokens[8:16]
['fittest', '.', 'Survival', 'of', 'the', 'most', 'adequate', ',']
>>> tokens[16:]
['maybe', '.']
>>> import numpy as np  # <1>
>>> vocab = sorted(set(tokens))  # <2>
>>> ' '.join(vocab[:12])  # <3>
", . Survival There's adequate as fittest maybe most no of such"
>>> num_tokens = len(tokens)
>>> num_tokens
18
>>> vocab_size = len(vocab)
>>> vocab_size
15
texts
>>> import re
>>> pattern = r'\w+(?:\'\w+)?|[^\w\s]'  # <1>
>>> texts = [text]
>>> texts.append("There's no such thing as survival of the fittest. "
...              "Survival of the most adequate, maybe.")
"""

text = ("Trust me, though, the words were on their way, and when "
        "they arrived, Liesel would hold them in her hands like "
        "the clouds, and she would wring them out, like the rain.")
texts = [text]
texts.append("There's no such thing as survival of the fittest."
             "Survival of the most adequate, maybe.")
spacy.cli.download('en_core_web_sm')
spacy.cli.download('en_core_web_md')
spacy.cli.download('en_core_web_lg')
nlp = spacy.load('en_core_web_md')

docs = [nlp(txt) for txt in texts]
doc = docs[-1]
sentence_spans = list(doc.sents)
sentences = list(doc.sents)
svg = displacy.render(sentences[-1], style="dep")
with open('manuscript/images/ch02/survival-of-adequate-sentence-diagram.svg', 'w') as f:
    f.write(html)
html = displacy.render(sentences[-1], style="dep", page=True)
with open('manuscript/images/ch02/survival-of-adequate-sentence-diagram.html', 'w') as f:
    f.write(html)



from manim import *


class GraphAutoPosition(Scene):
    def construct(self):
        vertices = [1, 2, 3, 4, 5, 6, 7, 8]
        edges = [(1, 7), (1, 8), (2, 3), (2, 4), (2, 5),
                 (2, 8), (3, 4), (6, 1), (6, 2),
                 (6, 3), (7, 2), (7, 4)]
        autolayouts = ["spring", "circular", "kamada_kawai",
                       "planar", "random", "shell",
                       "spectral", "spiral"]
        graphs = [Graph(vertices, edges, layout=lt).scale(0.5)
                  for lt in autolayouts]
        r1 = VGroup(*graphs[:3]).arrange()
        r2 = VGroup(*graphs[3:6]).arrange()
        r3 = VGroup(*graphs[6:]).arrange()
        self.add(VGroup(r1, r2, r3).arrange(direction=DOWN))




reviews = [
    "I really love your chocolate ice cream.",
    "The vanilla ice cream was delicious.",
    "I like mint chocolate chip ice cream.",
    "Your chocolate ice cream is sick.",
    "I'm lactose intollerant. Ice cream makes me ill."
]

# [source,python]
# ----
good_words = [
    "good", "delicious", "great", "like", "love",
    "dank", "dope", "sick"]
ok_words = [
    "ok", "decent", "ice cream", "chocolate", "vanilla"]
bad_words = [
    "bad", "ill", "sick", "messy", "bland", "nasty"]
vocab = good_words + ok_words + bad_words
# ----


def count_keywords(sentence, keywords):
    keyword_counts = dict(zip(keywords, [0] * len(keywords)))
    for kw in keywords:
        if kw in sentence:
            keyword_counts[kw] += 1
    return keyword_counts


count_keywords(reviews[0], vocab)



import re
from pathlib import Path
import logging
import pkg_resources

log = logging.getLogger(__name__)


PKG_DIR = Path(__file__).absolute().resolve().parent
PKG_NAME = PKG_DIR.name
SRC_DIR = PKG_DIR.parent
REPO_DIR = SRC_DIR.parent
MANUSCRIPT_DIR = REPO_DIR.parent / 'nlpia-manuscript' / 'manuscript'
IMAGES_DIR = MANUSCRIPT_DIR / 'images'
ADOC_DIR = MANUSCRIPT_DIR / 'adoc'


def get_version():
    """ Look within setup.cfg for version = ... and within setup.py for __version__ = """
    version = '0.0.0'
    try:
        return pkg_resources.get_distribution(PKG_NAME)
    except Exception as e:
        log.error(e)
        log.warning(f"Unable to find {PKG_NAME} version so using {version}")
    return version

    # setup.cfg will not exist if package install in site-packages
    with (REPO_DIR / 'setup.cfg').open() as fin:
        for line in fin:
            matched = re.match(r'\s*version\s*=\s*([.0-9abrc])\b', line)
            if matched:
                return (matched.groups()[-1] or '').strip()


__version__ = get_version()

HOME_DIR = Path.home().resolve().absolute()
DATA_DIR_NAME = '.nlpia2-data'
DATA_DIR = PKG_DIR / DATA_DIR_NAME
if not DATA_DIR.is_dir():
    DATA_DIR = REPO_DIR / DATA_DIR_NAME
if not DATA_DIR.is_dir():
    DATA_DIR = HOME_DIR / DATA_DIR_NAME
    # try/except this and use tempfiles python module as backup
    DATA_DIR.mkdir(parents=True, exist_ok=True)

# canonical data directory to share data between nlpia2 installations
HOME_DATA_DIR = HOME_DIR / DATA_DIR_NAME
if not HOME_DATA_DIR.is_dir():
    HOME_DATA_DIR.mkdir(parents=True, exist_ok=True)

REPO_DATA_DIR, DATA_DIR = DATA_DIR, HOME_DATA_DIR

# DONE: create nlpia2/init.py
# DONE: add maybe_download() to init.py
# TODO: all required data files up to chapter07
# TODO: add list of all required data files to init.py
# TODO: ensure all files are in HOME_DATA_DIR (DATA_DIR is just a subset)
# TODO: move DATA_DIR constant to data.py
# DATA_FILENAMES = dict(
#     DATA_DIR
# )



# ch02.ipy

"""
>>> text = ("Trust me, though, the words were on their way, and when "
...         "they arrived, Liesel would hold them in her hands like "
...         "the clouds, and she would wring them out, like the rain.")
>>> tokens = text.split()
>>> tokens[:9]
['Trust',
 'me,',
 'though,',
 'the',
 'words',
 'were',
 'on',
 'their',
 'way,']
"""

# NLPIA IDEA: sentence rewriting for data augmentation:
#       replace all re.replace('(,\ and )([c])', text, ['. \2'.upper(), '. And \2', '. In addition \2', '. Also ', '. Furthermore '])
#       obviously '\2' upper needs to be find_all() and manually replace
# NLPIA IDEA: generate multiple versions of a sentence and select the one that meets your criteria (brevity, clarity, reading score, style)


r"""
[source,python]
----
>>> import numpy as np  # <1>
>>> vocab = sorted(set(tokens))  # <2>
>>> ' '.join(vocab[:12])  # <3>
", . Survival There's adequate as fittest maybe most no of such"
>>> num_tokens = len(tokens)
>>> num_tokens
18
>>> vocab_size = len(vocab)
>>> vocab_size
15
----
<1> `str.split()` is your quick-and-dirty tokenizer.
<2> Coercing the `list` into a `set` ensures that your vocabulary contains only *unique* tokens that you want to keep track of.
<3> Sorted lexographically (lexically) so punctuation comes before letters, and capital letters come before lowercase letters.
"""
import numpy as np  # <1>
vocab = sorted(set(tokens))  # <2>
' '.join(vocab[:12])  # <3>
# ", . Survival There's adequate as fittest maybe most no of such"
num_tokens = len(tokens)
num_tokens
# 18
vocab_size = len(vocab)
vocab_size
# 15

r"""
>>> import re
>>> pattern = r'\w+(?:\'\w+)?|[^\w\s]'  # <1>
>>> text = ("There's no such thing as survival of the fittest."
...         "Survival of the most adequate, maybe.")
>>> tokens = list(re.findall(pattern, text))

>>> tokens[:8]
["There's", 'no', 'such', 'thing', 'as', 'survival', 'of', 'the']

>>> tokens[8:16]
['fittest', '.', 'Survival', 'of', 'the', 'most', 'adequate', ',']

>>> tokens[16:]
['maybe', '.']
"""


r"""
>>> df_onehot = pd.DataFrame(onehot_vectors, columns=vocab)
>>> df_onehot.iloc[:,:8]
    ,  .  Survival  There's  adequate  as  fittest  maybe
0   0  0         0        1         0   0        0      0
1   0  0         0        0         0   0        0      0
2   0  0         0        0         0   0        0      0
3   0  0         0        0         0   0        0      0
4   0  0         0        0         0   1        0      0
5   0  0         0        0         0   0        0      0
6   0  0         0        0         0   0        0      0
7   0  0         0        0         0   0        0      0
8   0  0         0        0         0   0        1      0
9   0  1         0        0         0   0        0      0
10  0  0         1        0         0   0        0      0
11  0  0         0        0         0   0        0      0
12  0  0         0        0         0   0        0      0
13  0  0         0        0         0   0        0      0
14  0  0         0        0         1   0        0      0
15  1  0         0        0         0   0        0      0
16  0  0         0        0         0   0        0      1
17  0  1         0        0         0   0        0      0
"""


r"""
>>> import re
>>> pattern = r'\w+(?:\'\w+)?|[^\w\s]'  # <1>
>>> text = ("There's no such thing as survival of the fittest."
...         "Survival of the most adequate, maybe.")
>>> tokens = list(re.findall(pattern, text))

>>> tokens[:8]
["There's", 'no', 'such', 'thing', 'as', 'survival', 'of', 'the']
>>> tokens[8:16]
['fittest', '.', 'Survival', 'of', 'the', 'most', 'adequate', ',']
>>> tokens[16:]
['maybe', '.']
"""
import re

r"""
>>> pattern = r'\w+(?:\'\w+)?|[^\w\s]'  # <1>
>>> text = ("There's no such thing as survival of the fittest."
            "Survival of the most adequate, maybe.")
>>> tokens = re.findall(pattern, text)
>>> tokens
["There's", 'no', 'such', 'thing', 'as', 'survival', ...]
"""


def tokenize(text, pattern=r'\w+(?:\'\w+)?|[^\w\s]'):
    r""" Split English text into words, ignoring only 1 internal punctuation"

    returns list(re.findall(r'\w+(?:\'\w+)?|[^\w\s]', text))
    """
    return list(re.findall(pattern, text))



#----------------------------------------------------------------------------------
# preprocessing.py: Utils for preprocessing text
#----------------------------------------------------------------------------------

import glob
import json
import os
import re

from flashtext.keywords import KeywordProcessor


class TextPreprocessor(object):
    
    spaces = ['\u200b', '\u200e', '\u202a', '\u202c', '\ufeff', '\uf0d8', '\u2061', '\x10', '\x7f', 
                  '\x9d', '\xad', '\xa0']
    
    
    punct = [',', '.', '"', ':', ')', '(', '-', '!', '?', '|', ';', "'", '$', '&', '/', '[', ']', '>', 
             '%', '=', '#', '*', '+', '\\', '•', '~', '@', '£', '·', '_', '{', '}', '©', '^', '®', '`', 
             '<', '→', '°', '€', '™', '›', '♥', '←', '×', '§', '″', '′', 'Â', '█', '½', 'à', '…', '“', '★', 
             '”', '–', '●', 'â', '►', '−', '¢', '²', '¬', '░', '¶', '↑', '±', '¿', '▾', '═', '¦', '║', '―',
             '¥', '▓', '—', '‹', '─', '▒', '：', '¼', '⊕', '▼', '▪', '†', '■', '’', '▀', '¨', '▄', '♫', '☆',
             'é', '¯', '♦', '¤', '▲', 'è', '¸', '¾', 'Ã', '⋅', '‘', '∞', '∙', '）', '↓', '、', '│', '（', '»',
             '，', '♪', '╩', '╚', '³', '・', '╦', '╣', '╔', '╗', '▬', '❤', 'ï', 'Ø', '¹', '≤', '‡', '√', 'θ',
             '₹', 'º', 'β', '«', '´', '₤', '¡', '÷', '∅', 'α', 'π']

    
    def __init__(self):
        # Load json resources in this module (preprocessing)
        # into a flashtext KeywordProcessor instance. 
        self.kwp = self._load_resources()
        
        self.kw_replace_errors = []
        
    def _load_resources(self):
        text_replacement_dicts = []
        searchpath = os.path.join(os.path.dirname(__file__), 'json/*.json')
        files = glob.glob(searchpath)
        for filename in files:
            try:
                with open(filename) as f:  
                    data = json.load(f)
                    text_replacement_dicts.append(data)
                    print('loaded', filename)
            except:
                print('error: could not load', filename)
                   
        kwp = KeywordProcessor()
        for d in text_replacement_dicts:
            for key, value in d.items():
                kwp.add_keyword(key, value)
        
        return kwp
    
    def _replace_keywords(self, text):
        try:
            return self.kwp.replace_keywords(text)
        except:
            # flashtext hits list index out of bounds error sometimes.
            # save the text to try figure what went wrong
            self.kw_replace_errors.append(text)
        
        return text

    def preprocess(self, text):
        text = self._remove_space(text)
        text = self._replace_special_punctuation(text)
        text = self._clean_numbers(text)
        text = self._replace_keywords(text)
        text = self._remove_space(text) # remove extra spaces that might have been introduced
                
        return text
        
   
    def _remove_space(self, text):
        """
        Remove extra spaces and ending space if any
        """
        for space in TextPreprocessor.spaces:
            if space in text:
                text = text.replace(space, ' ')
        text = text.strip()
        text = re.sub('\s+', ' ', text)
        
        return text
    
    def _replace_special_punctuation(self, text):
        special_punc_mappings = {"—": "-", "–": "-", "_": "-", '”': '"', "″": '"', '“': '"', '•': '.', '−': '-',
                         "’": "'", "‘": "'", "´": "'", "`": "'", '\u200b': ' ', '\xa0': ' ','،':'','„':'',
                         '…': ' ... ', '\ufeff': ''}
        
        for punc in special_punc_mappings:
            if punc in text:
                text = text.replace(punc, special_punc_mappings[punc])
        return text
    
    def _clean_numbers(self, text):
        text = re.sub(r'(\d+)([a-zA-Z])', '\g<1> \g<2>', text)
        text = re.sub(r'(\d+) (th|st|nd|rd) ', '\g<1>\g<2> ', text)
        text = re.sub(r'(\d+),(\d+)', '\g<1>\g<2>', text)
        text = re.sub(r'(\d+)(e)(\d+)','\g<1> \g<3>', text)

        return text


from spacy import cli
cli.download('en_core_web_md')
from qary.etl.wikicrawl import *
pages = {'Abundance (economics)': None}
pages2 = walk_wikipedia(pages=pages, depth=1)
pages2
pages2['Abundance (economics)'].content
pages2['Abundance (economics)'][-1]
pages
pages2 = walk_wikipedia(pages=pages, depth=2)
pages
pages2['Abundance (economics)'][-1].content
print(pages2['Abundance (economics)'][-1].content)
print(pages2['Abundance (economics)'][-1].links)
print(pages2['Abundance (economics)'][-1].__dict__)
meili = CLIENT.index('wikipedia')
meili.add_documents_json([pages[t][-1].__dict__ for t in pages])
meili.get_task(0)
meili.get_task?
meili.get_task(2)
meili.search('prosocial')
ans = meili.search('prosocial')
ans.keys()
ans['hits']
ans.keys()
ans['nbHits']
ans['query']
ans['limit']
ans['processingTime']
ans['processingTimeMs']
ans['hits'].keys()
len(ans['hits'])
ans['hits'][0].keys()
ans['hits'][0]['title']
ans = meili.search('social')
len(ans['hits'])
pages['Ecological economics']
pages['Ecological economics'][-1]
pages['Ecological economics'][-1].content
'prosocial' in pages['Ecological economics'][-1].content.lower()
'social' in pages['Ecological economics'][-1].content.lower()
from nessvec.indexers import Index
from nessvec.files import load_embeddings
load_embeddings('word2vec')
df = _
df.loc['prosocial']
w2vi = Index(df.iloc[:500_000])
w2vi.get_nearest?
w2vi.get_nearest('prosocial')
w2vi.get_nearest(df.loc['prosocial'].values)
w2vi.get_nearest??
w2vi.get_nearest('prosocial')
w2vi = Index(df.iloc[:1_000_000])
w2vi.get_nearest('prosocial')
w2vi.get_nearest('social')
w2vi['prosocial']
w2vi.keys()
w2vi.get('prosocial')
w2vi.vocab
w2vi.vocab['prosocial']
w2vi.tok2id['prosocial']
pd.Index
import pandas as pd
pd.Index



import torch
import datasets

dataset = datasets.load_dataset('tweets_hate_speech_detection')


#######################################################
# tokenizer with integrated stemmer, and stopword filter

# For simplicity let's remove alphanumeric but keep @, #
import re  # noqa
import nltk  # noqa
from nltk.corpus import stopwords  # noqa
from nltk.stem.snowball import SnowballStemmer  # noqa

ss = SnowballStemmer('english')

try:
    sw = stopwords.words('english')
except LookupError:
    nltk.download("stopwords")
    sw = stopwords.words('english')


def split_tokens(row):                             # STEP
    row['all_tokens'] = [ss.stem(i) for i in       # 5
                         re.split(r" +",               # 3
                                  re.sub(r"[^a-z@# ]", "",      # 2
                                         row['tweet'].lower()))  # 1
                         if (i not in sw) and len(i)]  # 4
    return row


# tokenizer with integrated stemmer, and stopword filter
#######################################################

# Determine vocabulary so we can create mapping
dataset = dataset.map(split_tokens)


################################################################
# count token frequency so rare tokens can be filtered, also stem all words

from collections import Counter  # noqa

counts = Counter([i for s in dataset['train']['all_tokens'] for i in s])
counts = {k: v for k, v in counts.items() if v > 10}  # Filtering
vocab = list(counts.keys())
n_v = len(vocab)
id2tok = dict(enumerate(vocab))
tok2id = {token: id for id, token in id2tok.items()}

# Now correct tokens


def remove_rare_tokens(row):
    row['tokens'] = [t for t in row['all_tokens'] if t in vocab]
    return row


dataset = dataset.map(remove_rare_tokens)

# count token frequency so rare tokens can be filtered, also stem all words
################################################################


def windowizer(row, wsize=3):
    """
    Windowizer function for Word2Vec. Converts sentence to sliding-window
    pairs.
    """
    doc = row['tokens']
    wsize = 3
    out = []
    for i, wd in enumerate(doc):
        target = tok2id[wd]
        window = [
            i + j for j in range(-wsize, wsize + 1, 1)
            if (i + j >= 0) & (i + j < len(doc)) & (j != 0)
        ]

        out += [(target, tok2id[doc[w]]) for w in window]
    row['moving_window'] = out
    return row


dataset = dataset.map(windowizer)


from torch.utils.data import Dataset, DataLoader  # noqa


class Word2VecDataset(Dataset):
    """
    Takes a HuggingFace dataset as an input, to be used for a Word2Vec dataloader.
    """

    def __init__(self, dataset, vocab_size, wsize=3):
        self.dataset = dataset
        self.vocab_size = vocab_size
        self.data = [i for s in dataset['moving_window'] for i in s]

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]


BATCH_SIZE = 2**14
N_LOADER_PROCS = 10

dataloader = {}
for key in dataset.keys():
    dataloader = {
        key: DataLoader(
            Word2VecDataset(
                dataset[key],
                vocab_size=n_v),
            batch_size=BATCH_SIZE,
            shuffle=True,
            num_workers=N_LOADER_PROCS)
    }


################################################
# manual one-hot encoding (ohe)

from torch import nn  # noqa

size = 10
input = 3


def one_hot_encode(input, size):
    vec = torch.zeros(size).float()
    vec[input] = 1.0
    return vec


ohe = one_hot_encode(input, size)
linear_layer = nn.Linear(size, 1, bias=False)

# Set edge weights from 0 to 9 for easy reference
with torch.no_grad():
    linear_layer.weight = nn.Parameter(
        torch.arange(10, dtype=torch.float).reshape(linear_layer.weight.shape))

print(linear_layer.weight)
print(linear_layer(ohe))

# manual one-hot encoding (ohe)
################################################


################################################
# NN model

class Word2Vec(nn.Module):
    def __init__(self, vocab_size, embedding_size):
        super().__init__()
        self.embed = nn.Embedding(vocab_size, embedding_size)
        self.expand = nn.Linear(embedding_size, vocab_size, bias=False)

    def forward(self, input):
        # Encode input to lower-dimensional representation
        hidden = self.embed(input)
        # Expand hidden layer to predictions
        logits = self.expand(hidden)
        return logits


# Instantiate the model
EMBED_SIZE = 100  # Quite small, just for the tutorial
model = Word2Vec(n_v, EMBED_SIZE)

# Relevant if you have a GPU:
device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
model.to(device)

# Define training parameters
LR = 3e-4
EPOCHS = 10
loss_fn = nn.CrossEntropyLoss()
optimizer = torch.optim.AdamW(model.parameters(), lr=LR)


from tqdm import tqdm  # noqa

running_loss = []

progress_bar = tqdm(range(EPOCHS * len(dataloader['train'])))
for epoch in range(EPOCHS):
    epoch_loss = 0
    for center, context in dataloader['train']:
        center, context = center.to(device), context.to(device)
        optimizer.zero_grad()
        logits = model(input=context)
        loss = loss_fn(logits, center)
        epoch_loss += loss.item()
        loss.backward()
        optimizer.step()
        progress_bar.update(1)
    epoch_loss /= len(dataloader['train'])
    running_loss.append(epoch_loss)

import matplotlib.pyplot as plt  # noqa
plt.plot(running_loss)
plt.show(block=False)


# NN model
################################################


################################################
# Save (persist) NN model


from pathlib import Path  # noqa
from nessvec.constants import DATA_DIR  # noqa

model_dir = Path(DATA_DIR) / 'models'
model_dir.mkdir(exist_ok=True)

num_embeddings, embedding_dim = model.state_dict()["embed.weight"].shape
print(num_embeddings, embedding_dim == model.state_dict()["expand.weight"].shape)

import datetime  # noqa
now = datetime.datetime.now()

filename = f'Word2Vec-state_dict-{now.isoformat()[:13]}.pt'
filename = f'Word2Vec-state_dict-{num_embeddings}x{embedding_dim}+{embedding_dim}x{num_embeddings}.pt'
filename = f'Word2Vec-state_dict-{now.isoformat()[:13]}-loss_{loss.item()}.pt'

torch.save(model_dir / filename)

# NN model
################################################


################################################
# more verbose training


pbar = tqdm(range(EPOCHS * len(dataloader['train'])))
for epoch in range(EPOCHS):
    epoch_loss = 0
    for sample_num, (center, context) in enumerate(dataloader['train']):
        if not sample_num % len(dataloader['train']) == 2:
            print(center, context)
        center, context = center.to(device), context.to(device)
        if not sample_num % len(dataloader['train']) == 2:
            print(center, context)
        optimizer.zero_grad()
        logits = model(input=context)
        loss = loss_fn(logits, center)
        if not sample_num % 10000:
            # print(center, context)
            pbar.set_description(f'loss[{sample_num}] = {loss.item()}')
        epoch_loss += loss.item()
        loss.backward()
        optimizer.step()
        pbar.update(1)
    epoch_loss /= len(dataloader['train'])
    running_loss.append(epoch_loss)


# more verbose training
################################################


import seaborn as sns  # noqa
import matplotlib.pyplot as plt  # noqa
sns.set_theme()
plt.plot(running_loss)
plt.show(block=False)

sns.set_theme()
# sns.set_style('whitegrid')
# learning curve
plt.grid('on')
plt.xlabel('Epoch')
plt.ylabel('Loss')
plt.show()



# Dota 2 Hero character sheets

# Dota2 fandom have created some nice tabular data describing all the Dota 2 heros.


import pandas as pd
dfs = pd.read_html('https://dota2.fandom.com/wiki/Strength')
dfs[0]
#     Hero             Hero.1  ...  Base Intelligence  Intelligence Growth
# 0    NaN            Abaddon  ...                 18                  2.0
# 1    NaN          Alchemist  ...                 25                  1.8
# 2    NaN                Axe  ...                 18                  1.6
# 3    NaN        Beastmaster  ...                 16                  1.9
# ...
# 38   NaN               Tusk  ...                 18                  1.7
# 39   NaN          Underlord  ...                 17                  2.3
# 40   NaN            Undying  ...                 27                  2.8
# 41   NaN        Wraith King  ...                 18                  1.6

# [42 rows x 8 columns]


# Get tables for the other 2 hero categories.


categories = 'Strength Agility Intelligence'.split()
url = 'https://dota2.fandom.com/wiki'
df = pd.concat([pd.read_html(f'{url}/{c}')[0].fillna(c) for c in categories])


# What about some NL text describing each hero:


df = pd.read_html('https://dota2.fandom.com/wiki/Heroes/Introductions')[1]
attr = pd.concat([pd.read_html(f'{url}/{c}')[0].fillna(c) for c in categories])
attrs = attr
dfs = pd.read_html(f'{url}/Heroes/Introductions')[1:]
intros = pd.concat(dfs).set_index('Hero')
intros
#                                                      Description
# Hero
# Abaddon        Unlike the other scions of his house, Abaddon ...
# Alchemist      Having broken out of prison with his ogre acco...
# Axe            One by one the soldiers of the Red Mist fell, ...
# Beastmaster    Raised in the royal menagerie of Slom, Karroch...
# Brewmaster     For nine days Mangix fought the master of his ...
# ...                                                          ...
# Warlock        In the endless pursuit of rare texts, Demnok L...
# Windranger     Born in the midst of a ruinous tempest, Lyrale...
# Winter Wyvern  Frustrated with writer's block, Auroth, the El...
# Witch Doctor   A bizarre figure shambles across the land, sea...
# Zeus           Fond of mortal women, but even fonder of immor...

# [123 rows x 1 columns]


# Consolidated table of numerical attributes for each hero


attrs2 = pd.read_html(f'{url}/Table_of_hero_attributes')[0]
attrs2
#                    HERO   A  STR  STR+  STR30  ...  VS-N   TR  COL  HP/S  L
# 0               Abaddon NaN   22   2.6   97.4  ...   800  0.6   24  3.20  2
# 1             Alchemist NaN   25   2.9  109.1  ...   800  0.6   24  2.75  2
# 2    Ancient Apparition NaN   20   1.9   75.1  ...   800  0.6   24  2.25  0
# 3             Anti-Mage NaN   21   1.6   67.4  ...   800  0.6   24  2.35  2
# 4            Arc Warden NaN   22   2.6   97.4  ...   800  0.7   24  2.45  2
# ..                  ...  ..  ...   ...    ...  ...   ...  ...  ...   ... ..
# 118          Windranger NaN   18   3.0  105.0  ...   800  0.8   24  2.05  2
# 119       Winter Wyvern NaN   24   2.6   99.4  ...   800  0.6   24  2.65  2
# 120        Witch Doctor NaN   18   2.3   84.7  ...   800  0.6   24  2.05  2
# 121         Wraith King NaN   22   3.0  109.0  ...   800  0.6   24  2.45  2
# 122                Zeus NaN   19   2.1   79.9  ...   800  0.6   24  2.15  2

# [123 rows x 29 columns]


# Only thing missing now is the release date/year/version info:


releases = pd.read_html(f'{url}/Heroes_by_release')
releases = releases[0].set_index('Hero')


# Clean up the column names and indices:


attrs.columns = ['Dominant Attribute', 'Hero'] + list(attrs.columns[2:])
attrs = attrs.set_index('Hero')
attrs2 = attrs2.set_index('HERO')


# Make sure we have all the heroes' names in all the table indices:


names2 = set(attrs2.index)
names = set(attrs.index)
names - names2
# set()
names2 - names
# set()
names == names2
# True


# Now the `releases` DataFrame index:


releases.index = [i[0] for i in releases.index]
releases
#                     Dota 2         DotA Allstars Version
#                    Release  Status       Release Version
# Anti-Mage       2010-11-01   Alpha    2004-03-01    2.60
# Axe             2010-11-01   Alpha    2005-03-01    6.00
# Crystal Maiden  2010-11-01   Alpha    2004-02-03    0.60
# Dazzle          2010-11-01   Alpha    2005-03-01    6.00
# Drow Ranger     2010-11-01   Alpha    2004-02-03    0.60
# ...                    ...     ...           ...     ...
# Void Spirit     2019-11-26  Reborn             -    7.23
# Hoodwink        2020-12-17  Reborn             -    7.28
# Dawnbreaker     2021-04-09  Reborn             -    7.29
# Marci           2021-10-28  Reborn             -   7.30e
# Primal Beast    2022-02-23  Reborn             -    7.31

# [123 rows x 4 columns]
releases.columns
# MultiIndex([(       'Dota 2', 'Release'),
#             (       'Dota 2',  'Status'),
#             ('DotA Allstars', 'Release'),
#             (      'Version', 'Version')],
#            )
releases.columns.get_level_values(1)
# Index(['Release', 'Status', 'Release', 'Version'], dtype='object')
releases.columns = releases.columns.get_level_values(1)
# releases
#                    Release  Status     Release Version
# Anti-Mage       2010-11-01   Alpha  2004-03-01    2.60
# Axe             2010-11-01   Alpha  2005-03-01    6.00
# Crystal Maiden  2010-11-01   Alpha  2004-02-03    0.60
# Dazzle          2010-11-01   Alpha  2005-03-01    6.00
# Drow Ranger     2010-11-01   Alpha  2004-02-03    0.60
# ...                    ...     ...         ...     ...
# Void Spirit     2019-11-26  Reborn           -    7.23
# Hoodwink        2020-12-17  Reborn           -    7.28
# Dawnbreaker     2021-04-09  Reborn           -    7.29
# Marci           2021-10-28  Reborn           -   7.30e
# Primal Beast    2022-02-23  Reborn           -    7.31

# [123 rows x 4 columns]


# Confirm releases.index contains all the hero names:


names3 = set(releases.index)
names3
# {'Abaddon',
#  'Alchemist',
#  'Ancient Apparition',
#  'Anti-Mage',
# ...
#  'Witch Doctor',
#  'Wraith King',
#  'Zeus'}
names3 - names
# set()
names - names3
# set()
names == names3
# True


# Finally the intros:


intros.head()
#                                                    Description
# Hero
# Abaddon      Unlike the other scions of his house, Abaddon ...
# Alchemist    Having broken out of prison with his ogre acco...
# Axe          One by one the soldiers of the Red Mist fell, ...
# Beastmaster  Raised in the royal menagerie of Slom, Karroch...
# Brewmaster   For nine days Mangix fought the master of his ...
names4 = set(intros.index)
names == names4
# True


# Combine all the things:


df = pd.concat([attrs, attrs2, intros, releases], axis=1)
from nlpia2.constants import DATA_DIR  # noqa
from pathlib import Path  # noqa
df.to_csv(Path(DATA_DIR) / "dota2-heroes.csv")
df.head()
#             Dominant Attribute  Base Strength  Strength Growth  Base Agility  ...     Release  Status     Release  Version
# Abaddon               Strength             22              2.6            23  ...  2013-07-12    Gold  2005-11-04     6.20
# Alchemist             Strength             25              2.9            22  ...  2011-11-18    Beta  2006-08-25     6.36
# Axe                   Strength             25              3.4            20  ...  2010-11-01   Alpha  2005-03-01     6.00
# Beastmaster           Strength             23              2.9            18  ...  2011-05-20    Beta  2006-04-11     6.30
# Brewmaster            Strength             23              3.7            19  ...  2012-04-19    Beta  2004-04-30     5.10

# [5 rows x 40 columns]


# Magic to create the history logs used to build these docs:


# %hist -o -p -f /home/hobs/code/tangibleai/nlpia2/examples/ch06_dota2_wiki_heroes.hist.ipy.md
# %hist -o -p -f /home/hobs/code/tangibleai/nlpia2/ch06_dota2_wiki_heroes.hist.ipy.md



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nlpia2',
 'nlpia2.bonus',
 'nlpia2.bonus.ipython-sessions',
 'nlpia2.ch01',
 'nlpia2.ch02',
 'nlpia2.ch03',
 'nlpia2.ch04',
 'nlpia2.ch05',
 'nlpia2.ch06',
 'nlpia2.ch07',
 'nlpia2.ch07.cnn',
 'nlpia2.ch07.cnn.cnn',
 'nlpia2.ch08',
 'nlpia2.ch12',
 'nlpia2.ch12.inc',
 'nlpia2.ch12.inc.flashtext',
 'nlpia2.ch12.inc.preprocessing',
 'nlpia2.etl',
 'nlpia2.first-edition',
 'nlpia2.pug',
 'nlpia2.review_questions',
 'nlpia2.scripts',
 'nlpia2.unused']

package_data = \
{'': ['*'],
 'nlpia2.ch05': ['.ipynb_checkpoints/*'],
 'nlpia2.ch06': ['.ipynb_checkpoints/*'],
 'nlpia2.ch07.cnn': ['data/*'],
 'nlpia2.ch12.inc.preprocessing': ['json/*']}

install_requires = \
['ConfigArgParse==1.5.3',
 'Cython>=0.29.28',
 'beautifulsoup4==4.11.1',
 'boto3==1.22.2',
 'botocore==1.25.3',
 'django>=4.0,<5.0',
 'edit-distance==1.0.4',
 'elasticsearch==7.17.3',
 'environment==1.0.0',
 'gitpython>=3.1.27,<4.0.0',
 'graphviz>=0.20,<0.21',
 'h5py>=3.6.0,<4.0.0',
 'html2text>=2020.1.16,<2021.0.0',
 'html5lib==1.1',
 'jedi==0.18.1',
 'jupyter-client>=7.3.0',
 'jupyter-console>=6.4.2',
 'jupyter==1.0.0',
 'lxml==4.6.3',
 'm2r==0.2.1',
 'matplotlib-inline==0.1.3',
 'matplotlib==3.5.1',
 'meilisearch==0.18.2',
 'mistune==0.8.4',
 'pandas==1.1.3',
 'pillow==9.1.0',
 'pip==22.0.4',
 'pronouncing==0.2.0',
 'psutil>=5.8.0,<6.0.0',
 'python-dotenv>=0.13.0,<0.14.0',
 'python-slugify==4.0.0',
 'pyyaml==6.0',
 'qary>=0.7,<0.8',
 'rapidfuzz==0.9.1',
 'recommonmark==0.6.0',
 'scikit-image>=0.17.2,<0.18.0',
 'scikit-learn>=1.0,<2.0',
 'scipy==1.8.0',
 'seaborn>=0.11.2,<0.12.0',
 'sentence_transformers>=2.2.0,<3.0.0',
 'spacy==3.2.4',
 'torch==1.11.0',
 'tox>=3.25.0,<4.0.0',
 'tqdm>=4.60.0,<5.0.0',
 'wikipedia>=1.4,<2.0']

entry_points = \
{'console_scripts': ['cnn = ch07.cnn.main']}

setup_kwargs = {
    'name': 'nlpia2',
    'version': '0.0.17',
    'description': 'Natural language processing utilities and examples for the book Natural Language Processing in Action (nlpia) 2nd Edition by Hobson Lane and Maria Dyshel.',
    'long_description': '# nlpia2\n\n<!-- [![PyPI version](https://img.shields.io/pypi/pyversions/nlpia2.svg)](https://pypi.org/project/nlpia2/)\n [![License](https://img.shields.io/pypi/l/qary.svg)](https://pypi.python.org/pypi/qary/)\n -->\n<!-- https://gitlab.com/username/userproject/badges/master/coverage.svg\n -->\n[![codecov](https://codecov.io/gl/tangibleai/nlpia2/branch/master/graph/badge.svg)](https://codecov.io/gl/tangibleai/nlpia2)\n[![GitLab CI](https://gitlab.com/tangibleai/nlpia2/badges/master/pipeline.svg)](https://gitlab.com/tangibleai/nlpia2/badges/master/pipeline.svg)\n\n\n\nOfficial code repository for the book [_Natural Language Processing in Action, 2nd Edition_](https://proai.org/nlpia2e) by Maria Dyshel and Hobson Lane at [Tangible AI](https://tangibleai.com). It would not have happened without the generous work of [contributing authors](AUTHORS.md) and prosocial AI developers.\n\nTo get the most of this repository, you need to do two things.\n\n1. **Clone the repository** to your local machine if you want to execute the code locally or want local access to the data (recommended).\n2. **Create an environment** that has all the helpful/needed modules for Natural Language Processing In Action, 2nd Edition.\n\n## Clone the Repository\n\nIf you\'re currently viewing this file on gitlab, and want in the future to access the data and code local to your machine, you may clone this repository to your local machine. Navigate to your preferred directory to house the local clone (for example, you local _git_ directory) and execute:\n\n`git clone git@gitlab.com:prosocialai/nlpia2`\n\n## Create a Conda Environment\n\nTo use the various packages in vogue with today\'s advanced NLP referenced in the NLPIA 2nd Edition book, such as PyTorch and SpaCy, you need to install them in a conda environment.  To avoid potential conflics of such packages and their dependencies with your other python projects, it is a good practice to create and activate a _new_ conda environment.\n\nHere\'s how we did that for this book.\n\n1. **Make sure you have Anaconda3 installed.** Make sure you can run conda from within a bash shell (terminal). The `conda --version` command should say something like \'`4.10.3`.\n\n2. **Update conda itself**. Keep current the `conda` package, which manages all other packages. Your base environment is most likely called _base_ so you can execute `conda update -n base -c defaults conda` to bring that package up to date.  Even if _base_ is not the activated environment at the moment, this command as presented will update the conda package in the _base_ environment. This way, next time you use the `conda` command, in any environment, the system will use the updated _conda_ package.\n\n3. **Create a new environment and install the variety of modules needed in NLPIA 2nd Edition.**\n\nThere are two ways to do that.  \n\n### Use the script already provided in the repository (_`nlpia2/src/nlpia2/scripts/conda_install.sh`_)\n\nIf you have cloned the repository, as instructed above, you already have a script that will do this work. From the directory housing the repository, run\n`cd nlpia2/src/nlpia2/scripts/` and from there run `bash conda_install.sh` \n\n### Or manually execute portions of the script as follows\n\nFirst, create a new environment (or activate it if it exists)\n\n```bash\n# create a new environment named "nlpia2" if one doesn\'t already exist:\nconda activate nlpia2 \\\n    || conda create -n nlpia2 -y \'python==3.9.7\' \\\n    && conda activate nlpia2\n```\n\nOnce that completes, install all of `nlpia2`\'s conda dependences if they aren\'t already installed:\n\n``` bash\nconda install -c defaults -c huggingface -c pytorch -c conda-forge -y \\\n    emoji \\\n    ffmpeg \\\n    glcontext \\\n    graphviz \\\n    huggingface_hub \\\n    jupyter \\\n    lxml \\\n    manimpango \\\n    nltk \\\n    pyglet \\\n    pylatex \\\n    pyrr \\\n    pyopengl \\\n    pytest \\\n    pytorch \\\n    regex \\\n    seaborn \\\n    scipy \\\n    scikit-learn \\\n    sentence-transformers \\\n    statsmodels \\\n    spacy \\\n    torchtext \\\n    transformers \\\n    wikipedia \\\n    xmltodict\n```\n\nFinally, install via pip any packages not available through conda channels.  In such scenarios it is generally a better practice to apply all pip installs after _all_ conda installs.  Furthermore, to ensure the pip installation is properly configured for the python version used in the conda environment, rather than use `pip` or `pip3`, activate the environment and invoke pip by using `python -m pip`.\n\n``` bash\nconda activate nlpia2\npython -m pip install manim manimgl\n```\n\n## Ready, Set, Go!\n\nCongratulations! You now have the nlpia2 repository cloned which gives you local access to all the data and scripts need in the NLPIA Second Edition book, and you have created a powerful environment to use.  When you\'re ready to type or execute code, check if this environment is activated. If not, activate by executing:\n\n`conda activate nlpia2`\n\nAnd off you go tackle some serious Natural Language Processing, in order to make the world a better place for all.\n\nRun a jupyter notebook server within docker:\n`jupyter-repo2docker --editable .`\n',
    'author': 'Hobson Lane',
    'author_email': 'hobson@tangibleai.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://proai.org',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.3,<3.10.0',
}


setup(**setup_kwargs)



from pathlib import Path

import bs4
import pandas as pd

DATA_DIR = Path('~/.nlpia2-data').expanduser().resolve().absolute()
ADOC_DIR = DATA_DIR / 'manuscript' / 'adoc'
HTML_DIR = DATA_DIR / 'manuscript' / 'html'

ADOC_DIR.mkdir(parents=True, exist_ok=True)
HTML_DIR.mkdir(parents=True, exist_ok=True)


if __name__ == '__main__':
    paths = list(HTML_DIR.glob('Chapter*.html'))
    lines = []
    for p in paths:
        chapter = bs4.BeautifulSoup(p.open()).get_text()
        chapter = [line if line else ' ' for line in chapter.splitlines()]
        lines.extend(
            list(
                zip(
                    [p.name] * len(chapter),
                    range(len(chapter)),
                    chapter
                )
            )
        )
    df = pd.DataFrame(lines, columns='filename chapter text'.split())
    # df = df.dropna()
    # print(df.head())

    csv_path = DATA_DIR / 'manuscript' / 'manuscript.adoc.html.text.csv'
    df.to_csv(csv_path, index=False)
    df = pd.read_csv(csv_path)
    print(df.head())
    # print(df.sample(10))
    # print(df.tail())



import json
from pathlib import Path

import meilisearch
import spacy


spacy_en = spacy.load("en_core_web_md")


ENV = {}
with Path('~/.ssh/qary/.env').expanduser().open() as fin:
    for line in fin:
        kv = [s.strip() for s in line.split('=')]
        if line.strip() and len(kv) == 2 and all([len(s) for s in kv]):
            ENV[kv[0].upper()] = kv[1]


ENV_PATH = Path('~').expanduser() / '.ssh' / 'qary' / '.env'
ENV_TEXT = ENV_PATH.open().read()
MEILI_MASTER_KEY = ENV_TEXT.split('=')[1].strip()
MEILI_URL = 'https://search.qary.ai'


def get_api_keys(url=MEILI_URL, master_key=MEILI_MASTER_KEY):
    client = meilisearch.Client(f'{url}', f'{master_key}')
    response = client.get_keys()
    return [k['key'] for k in response['results']]


API_KEYS = get_api_keys()
MEILI_API_KEY = API_KEYS[-1]


def get_meili_client(url=MEILI_URL, api_key=MEILI_API_KEY):
    return meilisearch.Client(f"{url}", api_key)


CLIENT = get_meili_client()


def get_indexes():
    return dict([(i.uid, i) for i in CLIENT.get_indexes()])


def tokenize_en(text):
    return [tok.text for tok in spacy_en.tokenizer(text)]


def upload_json(filepath='movies.json',
                api_key=MEILI_API_KEY,
                host=ENV.get('MEILI_URL', MEILI_URL),
                index_name=None,
                ):
    """ upload a json file to a new index on melisearch (search.qary.ai)

    curl -H "X-Meili-API-Key: $MEILI_API_KEY" -X POST $MEILI_URL'/indexes/movies/documents' --data @movies.json
    """

    json_file = open(filepath)
    data = json.load(json_file)
    if index_name is None:
        index_name = Path(filepath).with_suffix('').name
    print(f'index_name={index_name}')
    CLIENT.index(index_name).add_documents(data)


def upload_dicts(list_of_dicts, index_name='default'):
    CLIENT.index(index_name).add_documents(list_of_dicts)



import requests
from beautifulsoup4 import BeautifulSoup

sections = "Dialogue,Taglines,Cast,External links".split(',')

url = 'https://en.wikiquote.org/wiki/WarGames'
soup = BeautifulSoup(requests.get(url).content)
in_dialog = False
dialog_lines = []
for line in soup.text.splitlines():
    if line.startswith(sections[0] or in_dialog):
        in_dialog = True



# neither year nor len are statistically significant predictors of sex
from pathlib import Path
import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression  # , Lasso

DATA_DIR = Path('.nlpia2-data')
df = pd.read_csv(DATA_DIR / 'baby-names-region.csv.gz')

np.random.seed(451)
df = df.sample(10_000)
names = df['name'].unique()
istrain = np.random.rand(len(names)) < .9
istest = ~istrain

dfistrain = pd.DataFrame([istrain]).T
dfistrain.columns = ['istrain']
dfistrain.sum()
dfistrain['name'] = names
istrain = df.merge(dfistrain, on='name')['istrain']


# istrain = np.random.rand(len(df)) < .9
df['len'] = df['name'].str.len()
model = LogisticRegression(class_weight='balanced', max_iter=2000)
model.fit(df[['len', 'year']][istrain], df['sex'][istrain], sample_weight=df['count'][istrain])
model.score(df[['len', 'year']][istrain], df['sex'][istrain], sample_weight=df['count'][istrain])
model.score(df[['len', 'year']][~istrain], df['sex'][~istrain], sample_weight=df['count'][~istrain])


ournames = ['Kemal', 'Syndee', 'Hobson', 'Maria']
probas = model.predict_proba(vectorizer.transform(ournames))
df_probas = pd.DataFrame(probas, index=ournames, columns=model.classes_)

training_set_names = df['name'][istrain].unique()
for n in ournames:
    print(n, n in training_set_names)

X_ournames = pd.DataFrame.sparse.from_spmatrix(
    vectorizer.transform(ournames),
    index=ournames,
    columns=vectorizer.get_feature_names_out())
X_ournames
model.predict_proba(X_ournames)
"""
>>> pd.DataFrame(model.predict_proba(X_ournames),
...     columns=model.classes_,
...     index=ournames).round(4)
...
             F       M
Kemal   0.0085  0.9915
Syndee  0.9997  0.0003
Hobson  0.0038  0.9962
Maria   1.0000  0.0000
"""

y_test = df['sex'][~istrain]
y_test_pred = model.predict(df[['len', 'year']][~istrain])
df_plot = pd.DataFrame()
df_plot['female'] = y_test
df_plot['female_pred'] = y_test_pred
y_test_proba = model.predict_proba(df[['len', 'year']][~istrain])
df_plot['female_proba'] = y_test_proba[:, 1]
df_plot.sample(30)



import pandas as pd
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer

dfs = pd.read_html('http://www.sideroad.com/Business_Communication/politically-correct-language.html')
df = dfs[0]
df.columns = 'insensitive sensitive'.split()

texts = [str(s) for s in df['insensitive']]
vectorizer = CountVectorizer(min_df=1, max_df=.5, ngram_range=(1, 3), stop_words=None)
vectorizer.fit(texts)
# CountVectorizer(max_df=0.5, ngram_range=(1, 3))
vectorizer.transform(texts)
countvecs = vectorizer.transform(texts).todense()
countvecs
# matrix([[0, 0, 0, ..., 0, 0, 0],
#         [0, 0, 0, ..., 0, 0, 0],
#         [0, 0, 0, ..., 0, 0, 0],
#         ...,
#         [0, 0, 0, ..., 0, 0, 0],
#         [0, 0, 0, ..., 0, 0, 0],
#         [0, 0, 0, ..., 0, 0, 0]])
cv = pd.DataFrame(countvecs, columns=vectorizer.get_feature_names())
#     acting  acting blonde  acting like  acting like wild  adults  ...  wild  wild indians  woman  woman the  woman the wife
# 0        0              0            0                 0       0  ...     0             0      0          0               0
# 1        0              0            0                 0       0  ...     0             0      0          0               0
#
# [26 rows x 154 columns]
cv.T
#                   0   1   2   3   4   5   6   7   8   9   10  11  12  13  14  15  16  17  18  19  20  21  22  23  24  25
# acting             0   0   0   1   0   0   0   0   0   0   0   0   0   0   0   0   1   0   0   0   0   0   0   0   0   0
# acting blonde      0   0   0   0   0   0   0   0   0   0   0   0   0   0   0   0   1   0   0   0   0   0   0   0   0   0
#
# [154 rows x 26 columns]
qv = m[0]
# cv
#     acting  acting blonde  acting like  acting like wild  adults  and  and the  and the theater  ...  white  white lie  wife  wild  wild indians  woman  woman the  woman the wife
# 0        0              0            0                 0       0    0        0                0  ...      0          0     0     0             0      0          0               0
# 1        0              0            0                 0       0    0        0                0  ...      0          0     0     0             0      0          0               0
cv.dot(qv)
# 0     0
# 1     0
# 2     0
# 3     2
# 4     0
# 5     0
# 6     0
cv.dot(qv).sort_values(ascending=False)
# 3     2
# 16    1
# 0     0
# 14    0
query = 'wild acting'
m = vectorizer.transform([query]).toarray()
qv = m[0]
cv.dot(qv).sort_values(ascending=False).index[0]
# 3
texts[cv.dot(qv).sort_values(ascending=False).index[0]]
# 'Acting like wild Indians'
vectorizer = CountVectorizer(min_df=1, max_df=.5, ngram_range=(3, 3), stop_words=None, analyzer='char')
# CountVectorizer(analyzer='char', max_df=0.5, ngram_range=(3, 3))
trgcounts = pd.DataFrame(vectorizer.transform(texts).toarray(), columns=vectorizer.get_feature_names())
trgcounts.columns[0]
' "p'
vectorizer.transform([query])
# <1x334 sparse matrix of type '<class 'numpy.int64'>'
#     with 7 stored elements in Compressed Sparse Row format >
qv = vectorizer.transform([query]).toarray()[0]
trgcounts.dot(qv)
# 0     0
# 1     1
# 2     1
# 3     7
# 4     1
query = 'wild acting'
query = 'wld actng'
trgcounts.dot(qv)
# 0     0
# 1     1
# 2     1
# 3     7
# 4     1
# 5     0
qv = vectorizer.transform([query]).toarray()[0]
trgcounts.dot(qv)
# 0     0
# 1     0
# 2     0
# 3     2
# 4     0
# 5     0
query = 'wild actin'
qv = vectorizer.transform([query]).toarray()[0]
trgcounts.dot(qv)
# 0     0
# 1     0
# 2     0
# 3     6
# 4     0
# 5     0
trgcounts.sum()
#  "p    1
#  (w    5
#  a     1
#  ad    1
#  an    1
#       ..
# wor    1
# xed    1
# xis    1
# y o    1
# ys"    1
# Length: 334, dtype: int64
counts = trgcounts.sum()
counts.index
# Index([' "p', ' (w', ' a ', ' ad', ' an', ' ar', ' bl', ' bo', ' ch', ' co',
from collections import Counter
from collections import Counter
c = Counter()
for bg, num in zip(counts.index.str[:2], counts.values):
    c += Counter((bg, num))
c
# Counter({' "': 1,
#          1: 268,
#          ' (': 1,
#          5: 16,
counts.sum()
# 485
prob = pd.Series(c) / counts.sum()
seed='he'
for i in range(30):
    try:
        seed += prob.get(seed[-2:])
    except TypeError:
        break
prob.get('he')
0.010309278350515464
prob
#  "    0.002062
# 1     0.552577
#  (0.002062
# 5     0.032990
# [bigram for bigram, onegram in zip(*c.keys())]



import os
import graphviz
import shutil
from pathlib import Path

from nessvec.text import tokenize


CODE_DIR = Path(__file__).resolve().absolute()
for i in range(4):
    if CODE_DIR.name in ['code', 'nlpia2']:
        break
    # print(f"NOT code dir: {CODE_DIR}")
    CODE_DIR = CODE_DIR.parent

REPO_DIR = CODE_DIR.parent
for i in range(4):
    if REPO_DIR.name in ['nlpia-manuscript', 'nlpia2']:
        break
    # print(f"not repo dir: {REPO_DIR}")
    REPO_DIR = REPO_DIR.parent

HOME_CODE_DIR = REPO_DIR.parent.parent
print(HOME_CODE_DIR)
assert HOME_CODE_DIR.name == 'code'
MANUSCRIPT_DIR = HOME_CODE_DIR / 'tangibleai' / 'nlpia-manuscript' / 'manuscript'
assert MANUSCRIPT_DIR.is_dir()
IMAGE_DIR = MANUSCRIPT_DIR / 'images'
assert IMAGE_DIR.is_dir()
SCRIPT_WORKING_DIR = os.getcwd()

DEFAULT_SENTENCE = "Teaching is the highest form of understanding."


def draw_circle(sentence=DEFAULT_SENTENCE, name='sentence-circle-graphviz',
                class_name='Graph', engine='circo', center_name='attention',
                image_dir=IMAGE_DIR, image_subdir=''):
    """ Draw a sentence in a circle with connections between all words.
    Inputs:
      engine (https://graphviz.org/docs/layouts/):
        dot: hierarchies that look like org charts (best for Digraphs)
        neato: minimum energy layoud (Brownian motion? Lagrange functions?)
        circo: all nodes on edge of circle
        fdp: force-directed graph that reduces spring force instead of energy (like neato), order of "appearance" out from center (first nodes)
        osage: hierarchical clusters of subgraphs
        patchwork: squarified treemap layout
        sfdp: multiscale fdp for large graphs with many nodes
    """
    print(f'name: {name}')
    print(f'sentence: {sentence}')
    tokens = tokenize(sentence)
    print(f'tokens: {tokens}')
    print(f'image_dir: {image_dir}')
    dest = image_dir / Path(image_subdir) / (name + '.png')

    print('Destination circle network diagram path:')
    print(dest)

    g = getattr(graphviz, class_name)(name, engine=engine)  # dot neato fdp sfdp circo
    g.attr(rankdir='LR')
    g.attr('node', shape='box')
    g.node(center_name)
    g.attr('node', shape='plaintext')
    # g.edge('Text\n(natural language)', center_name)

    # g.save()

    g.render(filename=name, cleanup=1, view=0, format='png')

    try:
        dest.resolve().absolute().unlink()
    except FileNotFoundError:
        pass
    shutil.move(name + '.png', str(dest.resolve().absolute()))
    print(dest)
    return g

# g.render(filename=name, cleanup=1, view=0, format='pdf')
# g.render(filename=name, cleanup=1, view=0, format='svg')
# !firefox text-NLU-vector.svg


def draw_matrix(sentence=DEFAULT_SENTENCE, name='sentence-matrix', image_dir=IMAGE_DIR):
    print(f'name: {name}')
    print(f'sentence: {sentence}')
    tokens = tokenize(sentence)
    print(f'tokens: {tokens}')
    print(f'image_dir: {image_dir}')


if __name__ == '__main__':
    draw_circle()
    draw_matrix()
